
/*
*	$Id: thermgif.c,v 1.2 1996/11/12 06:40:15 manchek Exp $
*
*	Write a GIF file of a plot of the last week's temperature
*	Actually, it writes a PPM file unless it can fork ppmtogif...
*
*	Uses "tail" to get the end of a temperature logfile,
*	with lines in the format:
*		yr mo dy hr mn t
*	e.g.:
*		96 01 14 07 15 84
*
*	27 feb 1995 Robert Manchek manchek@cs.utk.edu
*
*	24 jan 1996 cleaned it up a bit
*	17 mar 1996 switch to floating-point temperatures
*/

/*
*  Copyright (C) 1995 by Robert Manchek.  manchek@CS.UTK.EDU.
*
*  Permission to use, copy, modify, and distribute this software
*  and its documentation for any purpose and without fee is hereby
*  granted, provided that this copyright notice appear in all
*  copies and that both that copyright notice and this permission
*  notice appear in supporting documentation.
*
*  This software is provided "as is" without express or implied
*  warranty.
*/

/*
*	$Log: thermgif.c,v $
*	Revision 1.2  1996/11/12 06:40:15  manchek
*	code cleanup
*
*	Revision 1.1  1996/11/12 06:29:14  manchek
*	Initial revision
*
*/

#ifndef	DOLOG
#define	DOLOG	0
#endif

#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#include <sys/time.h>
#if	DOLOG
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <syslog.h>
#endif
#include <imalloc.h>


#ifndef	BZERO
#define	BZERO(p,n)		bzero(p, n)
#endif

/*
*	defaults
*/

#ifndef	LOGFILE
#define	LOGFILE			"/usr/local/etc/temp.log"
#endif
#ifndef	TITLE1
#define	TITLE1			"Temperature Outside"
#endif
#ifndef	TITLE2
#define	TITLE2			"Current time: "
#endif
#ifndef	PPMTOGIF
#define	PPMTOGIF		"/usr/local/pbmplus/ppmtogif"
#endif
#ifndef	PPMQUANT
#define	PPMQUANT		"/usr/local/pbmplus/ppmquant"
#endif
#define	LABEL_HOW		'b'
#define	INCELS			0
#define	DEFAULTFGC		"#ffd700"
#define	DEFAULTAXC		"#a52a2a"
#define	DEFAULTBGC		"#202020"

#define	NOTEMP			999

#define	FONT_ASCENT		12
#define	FONT_DESCENT	3
#define	IMAGE_WIDTH		672
#define	IMAGE_HEIGHT	150

#ifndef	min
#define	min(a,b)	((a)<(b)?(a):(b))
#endif
#ifndef	max
#define	max(a,b)	((a)>(b)?(a):(b))
#endif

/* a temp sample */

struct tempdot {
	int yr;
	int mo;
	int da;
	int hr;
	int mn;
	float deg;
};

/* an image - cheap version of an X drawable,
   which is what this program did first */

struct img {
	int im_wd;
	int im_ht;
	char *im_data;		/* wd * ht * 3 */
};

struct colr {
	int r;
	int g;
	int b;
	int opac;
};


#ifdef _PT_
#undef _PT_
#endif
#ifdef __STDC__
#define _PT_(x) x
#else
#define _PT_(x) ()
#endif

int readppm _PT_(( char *name, FILE *ff, int *wdp, int *htp, char **datap ));
int writeppm _PT_(( char *name, FILE *ff, int wd, int ht, char *data ));

struct img * img_new _PT_(( int wd, int ht ));
int img_drawline _PT_(( struct img *imp, struct colr *cp, int x1, int y1, int x2, int y2 ));
int istic _PT_(( int hr, int mn ));
int draw_graph _PT_(( struct img *imp, int ovly ));
int iden _PT_(( int d ));
int main _PT_(( int argc, char **argv ));
int readem _PT_(( int count ));
int img_plot _PT_(( struct img *imp, struct colr *cp, int x, int y ));
int img_drawchar _PT_(( struct img *imp, struct colr *cp1, struct colr *cp2, int x, int y, char c ));
int img_drawstring _PT_(( struct img *imp, struct colr *cp1, struct colr *cp2, int x, int y, char *s, int l ));
int img_textwidth _PT_(( struct img *imp, char *s ));
int xdig _PT_(( char c ));
int parsecolor _PT_(( char *name, struct colr *col ));


/************
*  Globals  *
*           *
************/

static char rcsid[] = "$Id: thermgif.c,v 1.2 1996/11/12 06:40:15 manchek Exp $";
static struct tempdot *temps = 0;
static int ntemp = 0;
static int solid = 0;
static int label = LABEL_HOW;
static char *title = TITLE1;
static struct colr cols[3];
static char *logfile = LOGFILE;
static int cels = INCELS;


/*	img_new()
*
*	Create a new image with specified size
*/

struct img *
img_new(wd, ht)
	int wd;
	int ht;
{
	struct img *ip;

	if ((ip = TALLOC(1, struct img, 0))) {
		ip->im_wd = wd;
		ip->im_ht = ht;
		if ((ip->im_data = TALLOC(wd * ht * 3, char, 0))) {
			BZERO(ip->im_data, wd * ht * 3);

		} else {
			MYFREE(ip);
			ip = 0;
		}
	}
	return ip;
}


/*	img_drawline()
*
*	Draw a line in an image using Bresenham's algorithm.
*/

int
img_drawline(imp, cp, x1, y1, x2, y2)
	struct img *imp;
	struct colr *cp;
	int x1, y1;
	int x2, y2;
{
	int big, small;		/* dimensions of rectangle to draw across */
	int dx, dy;			/* delta x, y */
	int xsgn, ysgn;		/* x, y direction */
	int x, y;			/* position accumulator */
	int a;				/* accum for interpolating */
	int i;

	if ((dx = x2 - x1) < 0) {
		dx = -dx;
		xsgn = -1;
	} else
		xsgn = 1;
	if ((dy = y2 - y1) < 0) {
		dy = -dy;
		ysgn = -1;
	} else
		ysgn = 1;

	big = max(dx, dy);
	small = min(dx, dy);

	x = x1;
	y = y1;

	a = big / 2;	/* but with symmetrical lines */
	for (i = 0; i <= big; i++) {
		img_plot(imp, cp, x, y);
		if ((a += small) > big) {
			a -= big;
			if (dx < dy)
				x += xsgn;
			else
				y += ysgn;
		}
		if (dy > dx)
			y += ysgn;
		else
			x += xsgn;
	}
	return 0;
}


int
istic(hr, mn)
	int hr, mn;
{
	if (hr >= 9 && hr < 15)
		return mn < 45 ? 1 : 0;
	if (hr < 3 || hr >= 21)
		return mn < 15 ? 1 : 0;
	return mn < 30 ? 1 : 0;
}


/*	draw_graph()
*
*	Draw the whole temperature graph with labels and stuff
*/

int
draw_graph(imp, ovly)
	struct img *imp;			/* base image */
	int ovly;					/* is overlay (reduced graph in corner) */
{
	int wd, ht;					/* of image */
	int oxo, oyo, owd, oht;		/* whole overlay left, top, width, height */
	int gxo, gyo, gwd, ght;		/* graph left, top, width, height */
	int n;
	int i;
	int x, y, y0;
	int v1, v2;
	char buf[200];
	time_t now;
	int lhow;

	if (!temps)
		return 0;

	wd = imp->im_wd;
	ht = imp->im_ht;
	if (ovly) {
		oxo = wd / 10 - 5;
		oyo = 5;
		owd = (wd * 9) / 10;
		oht = ht / 2;

		for (y = 0; y < oht; y++)
			for (x = 0; x < owd; x++)
				img_plot(imp, &cols[2], x + oxo, y + oyo);

	} else {
		oxo = 0;
		oyo = 0;
		owd = wd;
		oht = ht;
	}

	img_drawline(imp, &cols[1], oxo, oyo, oxo + owd - 1, oyo);
	img_drawline(imp, &cols[1], oxo, oyo + oht - 1, oxo + owd - 1, oyo + oht - 1);
	img_drawline(imp, &cols[1], oxo, oyo + 1, oxo, oyo + oht - 2);
	img_drawline(imp, &cols[1], oxo + owd - 1, oyo + 1, oxo + owd - 1, oyo + oht - 2);

	lhow = ((label == 'l' || label == 'b') ? 1 : 0)
			+ ((label == 'r' || label == 'b') ? 2 : 0);

/*
	gyo = font_info->ascent / 2;
	ght = ht - font_info->ascent / 2 - font_info->descent - gyo;
	gxo = XTextWidth(font_info, "188", 3);
*/
	gyo = 3 + FONT_ASCENT / 2;
	ght = oht - (FONT_ASCENT + FONT_DESCENT) / 2 - gyo
		- (FONT_ASCENT + FONT_DESCENT) * 2 - 1;
	gyo += oyo;

	gxo = img_textwidth(imp, "188");
	switch (lhow) {
	case 1:
		gwd = owd - gxo;
		break;
	case 2:
		gwd = owd - gxo;
		gxo = 0;
		break;
	case 3:
		gwd = owd - 2 * gxo;
		break;
	case 0:
		gwd = owd;
		gxo = 0;
		break;
	}
	gxo += oxo;

	/* get min and max temperatures */

	n = min(gwd, ntemp);
	v1 = 200;
	v2 = -200;
	for (i = 0; i < n; i++) {
		if (temps[ntemp - n + i].deg != NOTEMP) {
			if (temps[ntemp - n + i].deg < v1)
				v1 = temps[ntemp - n + i].deg;
			if (temps[ntemp - n + i].deg > v2)
				v2 = temps[ntemp - n + i].deg + 1;
		}
	}
	if (v1 > v2)
		v1 = v2 = 0;
	if (v1 < 0)
		v1 = ((v1 - 9) / 10) * 10;
	else
		v1 = (v1 / 10) * 10;
	if (v2 < 0)
		v2 = (v2 / 10) * 10;
	else
		v2 = ((v2 + 9) / 10) * 10;
	if (v1 == v2)
		v2 += 10;
/*
	printf("%d to %d\n", v1, v2);
*/
	/* draw day scale */

	for (i = 0; i < n; i++) {
		if (temps[ntemp - n + i].mn < 15) {
			if (temps[ntemp - n + i].hr == 0) {
				x = gxo + gwd - n + i;
				img_drawline(imp, &cols[1], x, gyo,
						x, gyo + FONT_ASCENT + FONT_DESCENT + 1);
/*
				XDrawLine(xDisp, w, gcs[1], x, gyo,
						x, gyo + font_info->ascent + font_info->descent + 1);
*/
			}
			if (temps[ntemp - n + i].hr == 12) {
				sprintf(buf, "%d", temps[ntemp - n + i].da);
				x = gxo + gwd - n + i - img_textwidth(imp, buf) / 2;
				img_drawstring(imp, &cols[0], (struct colr *)0,
						x, gyo + FONT_ASCENT + 2,
						buf, strlen(buf));
/*
				XDrawString(xDisp, w, gcs[1],
						x, gyo + font_info->ascent + 2,
						buf, strlen(buf));
*/
			}
		}
	}


	/* draw scale */

	if (lhow)
		for (i = v1; i <= v2; i += 10) {
			sprintf(buf, "%d", i);
			y = gyo + ght - 1 - ((i - v1) * (ght - 1)) / (v2 - v1);
			img_drawline(imp, &cols[1], gxo, y, gxo + gwd - 1, y);
/*
			XDrawLine(xDisp, w, gcs[1], gxo, y, gxo + gwd - 1, y);
*/
			if (lhow & 1)
				img_drawstring(imp, &cols[0], (struct colr *)0,
						oxo + 2, y + FONT_ASCENT / 2,
						buf, strlen(buf));
/*
				XDrawString(xDisp, w, gcs[1],
						oxo + 2, y + font_info->ascent / 2,
						buf, strlen(buf));
*/
			if (lhow & 2)
				img_drawstring(imp, &cols[0], (struct colr *)0,
						gxo + gwd + 2, y + FONT_ASCENT / 2,
						buf, strlen(buf));
/*
				XDrawString(xDisp, w, gcs[1],
						gxo + gwd + 2, y + font_info->ascent / 2,
						buf, strlen(buf));
*/
		}

	/* draw plot and hour scale */

	y0 = -1;
	for (i = 0; i < n; i++) {
		x = gxo + gwd - n + i;
		if (temps[ntemp - n + i].deg != NOTEMP) {
			y = gyo + ght - 1
					- ((temps[ntemp - n + i].deg - v1) * (ght - 1)) / (v2 - v1);
			if (solid) {
				img_drawline(imp, &cols[0], x, y, x, ht - 1);
/*
				XDrawLine(xDisp, w, gcs[0], x, y, x, ht - 1);
*/
			} else {
				if (y0 == -1) {
					img_drawline(imp, &cols[0], x, y, x, y);
/*
					XDrawLine(xDisp, w, gcs[0], x, y, x, y);
*/

				} else {
					if (y < y0) {
						img_drawline(imp, &cols[0], x, y0, x, y);
/*
						XDrawLine(xDisp, w, gcs[0], x, y0, x, y);
*/
					} else {
						img_drawline(imp, &cols[0], x - 1, y0, x - 1, y);
						img_drawline(imp, &cols[0], x - 1, y, x, y);
/*
						XDrawLine(xDisp, w, gcs[0], x - 1, y0, x - 1, y);
						XDrawLine(xDisp, w, gcs[0], x - 1, y, x, y);
*/
					}
				}
			}
			y0 = y;
		} else
			y0 = -1;
		if (istic(temps[ntemp - n + i].hr, temps[ntemp - n + i].mn))
			img_drawline(imp, &cols[0], x, oyo + 1, x, oyo + 3);
/*
			XDrawLine(xDisp, w, gcs[0], x, oyo + 1, x, oyo + 3);
*/
	}

	if (title && title[0]) {
		img_drawstring(imp, &cols[0], (struct colr *)0,
				oxo + (owd - img_textwidth(imp, title)) / 2,
				oyo + oht - FONT_ASCENT - FONT_DESCENT * 2,
				title, strlen(title));
	}

	time(&now);
	strcpy(buf, TITLE2);
	strcat(buf, ctime(&now));
	if (buf[strlen(buf) - 1] == '\n')
		buf[strlen(buf) - 1] = 0;
	img_drawstring(imp, &cols[0], (struct colr *)0,
			oxo + (owd - img_textwidth(imp, buf)) / 2, oyo + oht - FONT_DESCENT,
			buf, strlen(buf));
	return 0;
}


#if	DOLOG
/*	iden()
*
*	syslog a line identifying the peer socket (like if we're started by inetd)
*/

int
iden(d)
	int d;
{
	struct sockaddr_in sin;
	int l = sizeof(sin);
	unsigned int a;
	struct hostent *he;

	if (getpeername(d, (struct sockaddr *)&sin, &l) == 0) {
		if ((he = gethostbyaddr((char *)&sin.sin_addr, sizeof(sin.sin_addr), AF_INET))) {
			syslog(LOG_INFO, "thermgif: from %s", he->h_name);
		} else {
			a = ntohl(sin.sin_addr.s_addr);
			syslog(LOG_INFO, "thermgif: from %d.%d.%d.%d",
					0xff & (a >> 24),
					0xff & (a >> 16),
					0xff & (a >> 8),
					0xff & a);
		}
	} else {
		syslog(LOG_INFO, "thermgif: getpeername: %m");
	}

	return 0;
}
#endif


int
main(argc, argv)
	int argc;
	char **argv;
{
	struct img *im;
	int pfd[2], p2fd[2];
	int i, j;
	int err = 0;
	char *p;
	int wd = IMAGE_WIDTH;
	int ht = IMAGE_HEIGHT;
	char *srcimage = 0;
	FILE *ff;

	parsecolor(DEFAULTFGC, &cols[0]);
	parsecolor(DEFAULTAXC, &cols[1]);
	parsecolor(DEFAULTBGC, &cols[2]);
cols[0].opac = 255;
cols[1].opac = 255;
cols[2].opac = 96;

	for (i = j = 1; i < argc; i++) {
		if (argv[i][0] == '-')
			for (p = argv[i] + 1; *p; p++) {
				switch (*p) {
				case 'l':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -l\n");
						err++;
					} else
						label = argv[i][0];
					break;

				case 'n':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -n\n");
						err++;
					} else
						logfile = argv[i];
					break;

				case 's':
					solid = 1;
					break;

				case 'c':
					cels = 1;
					break;

				case 'f':
					cels = 0;
					break;

				case 't':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -t\n");
						err++;
					} else
						title = argv[i];
					break;

				case 'b':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -b\n");
						err++;
					} else
						parsecolor(argv[i], &cols[2]);
					break;

				case 'g':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -g\n");
						err++;
					} else
						parsecolor(argv[i], &cols[0]);
					break;

				case 'a':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -a\n");
						err++;
					} else
						parsecolor(argv[i], &cols[1]);
					break;

				case 'w':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -w\n");
						err++;
					} else
						wd = atoi(argv[i]);
					break;

				case 'h':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -h\n");
						err++;
					} else
						ht = atoi(argv[i]);
					break;

				case 'i':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -h\n");
						err++;
					} else
						srcimage = argv[i];
					break;

				default:
					fprintf(stderr, "unknown switch: %c\n", *p);
					err++;
					break;
				}
			}

		else
			argv[j++] = argv[i];
	}
	argc = j;

	if (err) {
		fprintf(stderr, "usage: thermgif [ options ]\n");
		fprintf(stderr, "options: [-n logfile]\n");
		fprintf(stderr, "         [-c] display in celsius\n");
		fprintf(stderr, "         [-f] display in fahrenheit\n");
		fprintf(stderr, "         [-t title] graph title\n");
		fprintf(stderr, "         [-l lrbn] label temps on left, right, both, none\n");
		fprintf(stderr, "         [-s] fill area below graph\n");
		fprintf(stderr, "         [-w wd] image width\n");
		fprintf(stderr, "         [-h ht] image height\n");
		fprintf(stderr, "         [-b #xxx] background color\n");
		fprintf(stderr, "         [-g #xxx] graph color\n");
		fprintf(stderr, "         [-a #xxx] axis color\n");
		fprintf(stderr, "         [-i name/-] include source image\n");
		exit(1);
	}

	if (!logfile) {
		fprintf(stderr, "thermgif: no temp log file\n");
		exit(1);
	}

#if	DOLOG
	openlog("thermgif", LOG_PID, LOG_LOCAL0);
	iden(0);
#endif

	if (srcimage) {
		if (!strcmp(srcimage, "-"))
			ff = stdin;
		else if (!(ff = fopen(srcimage, "r"))) {
			fprintf(stderr, "%s: can't read\n", srcimage);
			exit(1);
		}
		im = TALLOC(1, struct img, 0);
		if (readppm(srcimage, ff, &im->im_wd, &im->im_ht, &im->im_data))
			exit(1);
		if (ff != stdin)
			fclose(ff);
/*
		fprintf(stderr, "read %d x %d base image\n", im->im_wd, im->im_ht);
*/

	} else {
		im = img_new(wd, ht);
	}

	readem(im->im_wd);

	draw_graph(im, (srcimage ? 1 : 0));

#ifdef	PPMTOGIF
	pipe(pfd);

	if (!fork()) {
		close(2);
		close(pfd[1]);
		dup2(pfd[0], 0);
		execlp(PPMTOGIF, PPMTOGIF, (char *)0);
/*
		execlp(PPMTOGIF, PPMTOGIF, "-interlace", (char *)0);
*/
		perror(PPMTOGIF);
		_exit(1);
	}

	close(pfd[0]);
#ifdef	PPMQUANT
	pipe(p2fd);
	if (!fork()) {
		close(2);
		close(p2fd[1]);
		dup2(p2fd[0], 0);
		dup2(pfd[1], 1);
		execlp(PPMQUANT, PPMQUANT, "128", (char *)0);
		_exit(1);

	}
	close(p2fd[0]);
	dup2(p2fd[1], 1);
	close(pfd[1]);

#else
	dup2(pfd[1], 1);
#endif

#endif

	fprintf(stdout, "P6\n%d %d\n255\n", im->im_wd, im->im_ht);
	fwrite(im->im_data, 1, im->im_wd * im->im_ht * 3, stdout);

#ifdef	PPMTOGIF
	fclose(stdout);
	wait((int *)0);
#ifdef	PPMQUANT
	wait((int *)0);
#endif
#endif

	exit(0);
}


/*	readem()
*
*	Read tail of temperature log file
*/

int
readem(count)
	int count;
{
	int stemp;
	int y, m, d, h, n;
	float t;
	int i;
	FILE *ff;
	char buf[200];

	if (temps) {
		stemp = ntemp;

	} else {
		stemp = 50;
		temps = TALLOC(stemp, struct tempdot, 0);
	}
	ntemp = 0;

	sprintf(buf, "tail -%d %s", count, logfile);
	ff = popen(buf, "r");
/*
	printf("readem() %s\n", buf);
*/
	while (fgets(buf, sizeof(buf)-1, ff)) {
		if ((i = sscanf(buf, " %d %d %d %d %d %f", &y, &m, &d, &h, &n, &t)) > 4)
		{
			if (ntemp == stemp) {
				stemp += stemp / 2;
				temps = TREALLOC(temps, stemp, struct tempdot);
			}
			temps[ntemp].yr = y;
			temps[ntemp].mo = m;
			temps[ntemp].da = d;
			temps[ntemp].hr = h;
			temps[ntemp].mn = n;
			if (i == 6) {
				if (cels)
					t = ((t - 32) * 5) / 9;
			} else
				t = NOTEMP;
			temps[ntemp++].deg = t;
		}
	}
	fclose(ff);
	(void)wait((int *)0);
	return 0;
}


unsigned char cdata[96][30] = {
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x90, 0x00, 0x90, 0x00, 0x90, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x48, 0x00, 0x48, 0x00,
  0xfc, 0x00, 0x48, 0x00, 0x48, 0x00, 0xfc, 0x00, 0x48, 0x00,
  0x48, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0xf8, 0x00,
  0x24, 0x00, 0x24, 0x00, 0xf8, 0x00, 0x20, 0x01, 0x20, 0x01,
  0xf8, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x08, 0x01, 0x94, 0x00, 0x94, 0x00,
  0x48, 0x00, 0x20, 0x00, 0x20, 0x00, 0x90, 0x00, 0x48, 0x01,
  0x48, 0x01, 0x84, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x24, 0x00, 0x24, 0x00,
  0x24, 0x00, 0x18, 0x00, 0x24, 0x00, 0x44, 0x01, 0x84, 0x00,
  0x44, 0x01, 0x38, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xe0, 0x00, 0x60, 0x00, 0x10, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x88, 0x00, 0x50, 0x00, 0xfc, 0x01, 0x50, 0x00, 0x88, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0xfc, 0x01, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x70, 0x00, 0x30, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x20, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x80, 0x00, 0x80, 0x00,
  0x40, 0x00, 0x20, 0x00, 0x20, 0x00, 0x10, 0x00, 0x08, 0x00,
  0x08, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x30, 0x00, 0x48, 0x00, 0x84, 0x00,
  0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00,
  0x48, 0x00, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x30, 0x00, 0x28, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x00, 0x01, 0x80, 0x00, 0x40, 0x00, 0x30, 0x00, 0x08, 0x00,
  0x04, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x00, 0x01, 0x80, 0x00,
  0x40, 0x00, 0xe0, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0xc0, 0x00, 0xa0, 0x00,
  0x90, 0x00, 0x88, 0x00, 0x84, 0x00, 0xfc, 0x01, 0x80, 0x00,
  0x80, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x04, 0x00, 0x04, 0x00,
  0xf4, 0x00, 0x0c, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf0, 0x00, 0x08, 0x00, 0x04, 0x00,
  0x04, 0x00, 0xf4, 0x00, 0x0c, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x80, 0x00, 0x40, 0x00, 0x20, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x84, 0x01, 0x78, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x80, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x20, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x20, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x20, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x70, 0x00, 0x30, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x40, 0x00, 0x20, 0x00,
  0x10, 0x00, 0x08, 0x00, 0x08, 0x00, 0x10, 0x00, 0x20, 0x00,
  0x40, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x10, 0x00, 0x20, 0x00,
  0x40, 0x00, 0x80, 0x00, 0x80, 0x00, 0x40, 0x00, 0x20, 0x00,
  0x10, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf0, 0x00, 0x08, 0x01, 0x08, 0x01,
  0x00, 0x01, 0x80, 0x00, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00,
  0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0xe4, 0x01, 0x14, 0x01, 0x94, 0x01, 0x64, 0x01, 0x04, 0x00,
  0x04, 0x00, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x50, 0x00, 0x88, 0x00,
  0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0xfc, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x08, 0x01, 0x08, 0x01,
  0x08, 0x01, 0xf8, 0x00, 0x08, 0x01, 0x08, 0x01, 0x08, 0x01,
  0x08, 0x01, 0xfc, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x00,
  0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x08, 0x01, 0x08, 0x01,
  0x08, 0x01, 0x08, 0x01, 0x08, 0x01, 0x08, 0x01, 0x08, 0x01,
  0x08, 0x01, 0xfc, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x08, 0x00, 0x08, 0x00,
  0x08, 0x00, 0x78, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00,
  0x08, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x08, 0x00, 0x08, 0x00,
  0x08, 0x00, 0x78, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00,
  0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x00,
  0x04, 0x00, 0x04, 0x00, 0xc4, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xfc, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xe0, 0x01, 0x80, 0x00, 0x80, 0x00,
  0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00,
  0x84, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x84, 0x00, 0x44, 0x00,
  0x24, 0x00, 0x1c, 0x00, 0x14, 0x00, 0x24, 0x00, 0x44, 0x00,
  0x84, 0x00, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x8c, 0x01,
  0x54, 0x01, 0x54, 0x01, 0x24, 0x01, 0x24, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x0c, 0x01,
  0x14, 0x01, 0x24, 0x01, 0x44, 0x01, 0x84, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xfc, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x24, 0x01,
  0x44, 0x01, 0xf8, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xfc, 0x00, 0x24, 0x00, 0x44, 0x00, 0x84, 0x00,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x00,
  0x04, 0x00, 0xf8, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x88, 0x00, 0x88, 0x00, 0x88, 0x00, 0x50, 0x00, 0x50, 0x00,
  0x50, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x24, 0x01, 0x24, 0x01, 0x24, 0x01, 0x24, 0x01,
  0x54, 0x01, 0x88, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x88, 0x00,
  0x50, 0x00, 0x20, 0x00, 0x20, 0x00, 0x50, 0x00, 0x88, 0x00,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x88, 0x00,
  0x50, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x00, 0x01, 0x80, 0x00,
  0x40, 0x00, 0x20, 0x00, 0x10, 0x00, 0x08, 0x00, 0x04, 0x00,
  0x04, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf0, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x10, 0x00, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x08, 0x00, 0x08, 0x00,
  0x10, 0x00, 0x20, 0x00, 0x20, 0x00, 0x40, 0x00, 0x80, 0x00,
  0x80, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x40, 0x00, 0x40, 0x00,
  0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00,
  0x40, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x50, 0x00, 0x88, 0x00,
  0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xfe, 0x01, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x30, 0x00, 0x40, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf8, 0x00, 0x00, 0x01, 0x00, 0x01, 0xf8, 0x01, 0x04, 0x01,
  0x84, 0x01, 0x78, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0xf4, 0x00, 0x0c, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x0c, 0x01, 0xf4, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf8, 0x00, 0x04, 0x01, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x00, 0xf8, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x78, 0x01, 0x84, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x84, 0x01, 0x78, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf8, 0x00, 0x04, 0x01, 0x04, 0x01, 0xfc, 0x01, 0x04, 0x00,
  0x04, 0x00, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xe0, 0x00, 0x10, 0x01, 0x10, 0x00,
  0x10, 0x00, 0x78, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x10, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x78, 0x01, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x78, 0x00,
  0x04, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01, 0xf8, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0xf4, 0x00, 0x0c, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00,
  0x30, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00,
  0xe0, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00,
  0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x84, 0x00, 0x78, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x01, 0xc4, 0x00, 0x34, 0x00, 0x0c, 0x00, 0x34, 0x00,
  0xc4, 0x00, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x30, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xdc, 0x00, 0x24, 0x01, 0x24, 0x01, 0x24, 0x01, 0x24, 0x01,
  0x24, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf4, 0x00, 0x0c, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf8, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf4, 0x00, 0x0c, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x0c, 0x01, 0xf4, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x78, 0x01, 0x84, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x84, 0x01, 0x78, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xe4, 0x00, 0x18, 0x01, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00,
  0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf8, 0x00, 0x04, 0x01, 0x04, 0x00, 0xf8, 0x00, 0x00, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x78, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x10, 0x01, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00,
  0x84, 0x00, 0x78, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x04, 0x01, 0x04, 0x01, 0x88, 0x00, 0x88, 0x00, 0x50, 0x00,
  0x50, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x04, 0x01, 0x04, 0x01, 0x24, 0x01, 0x24, 0x01, 0x24, 0x01,
  0x54, 0x01, 0x88, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x04, 0x01, 0x88, 0x00, 0x50, 0x00, 0x20, 0x00, 0x50, 0x00,
  0x88, 0x00, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00,
  0xc4, 0x00, 0xb8, 0x00, 0x80, 0x00, 0x80, 0x00, 0x78, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xfc, 0x01, 0x80, 0x00, 0x40, 0x00, 0x20, 0x00, 0x10, 0x00,
  0x08, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xc0, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x10, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x40, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x18, 0x01, 0x24, 0x01, 0xc4, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};


/*	img_plot()
*
*	Plot a pixel in an image
*/

int
img_plot(imp, cp, x, y)
	struct img *imp;
	struct colr *cp;
	int x, y;
{
	char *p;

	if (x < 0 || x >= imp->im_wd || y < 0 || y >= imp->im_ht) {
/*
		fprintf(stderr, "img_plot() out of range %d, %d\n", x, y);
		abort();
*/
		return 0;
	}
	p = &imp->im_data[(x + imp->im_wd * y) * 3];
	*p = ((0xff & *p) * (255 - cp->opac) + (0xff & cp->r) * cp->opac) / 255;
	p++;
	*p = ((0xff & *p) * (255 - cp->opac) + (0xff & cp->g) * cp->opac) / 255;
	p++;
	*p = ((0xff & *p) * (255 - cp->opac) + (0xff & cp->b) * cp->opac) / 255;
	p++;
/*
	*p++ = cp->r;
	*p++ = cp->g;
	*p = cp->b;
*/
	return 0;
}


/*	img_drawchar()
*
*	Draw a character in an image
*/

int
img_drawchar(imp, cp1, cp2, x, y, c)
	struct img *imp;
	struct colr *cp1, *cp2;
	int x, y;
	char c;
{
	int x1, y1;
	unsigned char *p;
	struct colr *cp;

	c &= 127;
	if (c < 32)
		c = ' ';
	p = cdata[c - ' '];
	for (y1 = 0; y1 < 15; y1++)
		for (x1 = 0; x1 < 11; x1++) {
			if (cp = (p[y1 * 2 + x1 / 8] & (1 << (x1 & 7))) ? cp1 : cp2)
				img_plot(imp, cp, x + x1, y - FONT_ASCENT + y1);
		}
	return 0;
}


/*	img_drawstring()
*
*	Draw a string in an image
*/

int
img_drawstring(imp, cp1, cp2, x, y, s, l)
	struct img *imp;
	struct colr *cp1, *cp2;
	int x, y;
	char *s;
	int l;
{
	while (*s) {
		img_drawchar(imp, cp1, cp2, x, y, *s);
		x += 11;
		s++;
	}
	return 0;
}


/*	img_textwidth()
*
*	Determine pixel width of text in image
*/

int
img_textwidth(imp, s)
	struct img *imp;
	char *s;
{
	return strlen(s) * 11;
}


/*	xdig()
*
*	Return value of hexadecimal character
*/

int
xdig(c)
	char c;
{
	return (c >= '0' && c <= '9') ? c - '0'
			: ((c >= 'A' && c <= 'F') ? c - 'A' + 10
			: ((c >= 'a' && c <= 'f') ? c - 'a' + 10
			: 0));
}


/*	parsecolor()
*
*	Parse a color name (in format #xxx or #xxxxxx)
*/

int
parsecolor(name, col)
	char *name;
	struct colr *col;
{
	if (name[0] != '#') {
		fprintf(stderr, "can't deal with color names: %s\n", name);
		return 1;
	}
	name++;
	switch (strlen(name)) {
		case 3:
			col->r = xdig(name[0]) * 16;
			col->g = xdig(name[1]) * 16;
			col->b = xdig(name[2]) * 16;
			break;

		case 6:
			col->r = xdig(name[0]) * 16 + xdig(name[1]);
			col->g = xdig(name[2]) * 16 + xdig(name[3]);
			col->b = xdig(name[4]) * 16 + xdig(name[5]);
			break;

		default:
			fprintf(stderr, "can't convert color: %s\n", --name);
			return 1;
			break;
	}
	return 0;
}


