
/*
*	$Id: draw.c,v 1.5 1998/02/20 01:00:16 manchek Exp $
*
*	Some stuff to scribble in images.
*	11 Nov 1996  Robert Manchek
*/

/*
*  Copyright (C) 1995, 1996 by Robert Manchek.  (manchek@spiderplant.com)
*
*  Permission to use, copy, modify, and distribute this software
*  and its documentation for any purpose and without fee is hereby
*  granted, provided that this copyright notice appear in all
*  copies and that both that copyright notice and this permission
*  notice appear in supporting documentation.
*
*  This software is provided "as is" without express or implied
*  warranty.
*/

/*
*	$Log: draw.c,v $
*	Revision 1.5  1998/02/20 01:00:16  manchek
*	added length param to img_textwidth
*
*	Revision 1.4  1996/11/21 04:23:03  manchek
*	switched to bfunc.h
*
*	Revision 1.3  1996/11/12 06:12:12  manchek
*	code cleanup
*
*	Revision 1.2  1996/11/12 06:04:32  manchek
*	changed font table to unsigned char
*
*	Revision 1.1  1996/11/12 05:19:19  manchek
*	Initial revision
*
*/

#include <stdlib.h>
#include <stdio.h>
#ifdef	SYSVSTR
#include <string.h>
#else
#include <strings.h>
#endif
#include <ctype.h>
#include <math.h>
#include <imalloc.h>
#include "draw.h"
#include "bfunc.h"

#define	FONT_ASCENT		12
#define	FONT_DESCENT	3

#ifndef	min
#define	min(a,b)	((a)<(b)?(a):(b))
#endif
#ifndef	max
#define	max(a,b)	((a)>(b)?(a):(b))
#endif

int readppm _PT_(( char *name, FILE *ff, int *wdp, int *htp, char **datap ));
int writeppm _PT_(( char *name, FILE *ff, int wd, int ht, char *data ));

/************
*  Globals  *
*           *
************/

static char rcsid[] = "$Id: draw.c,v 1.5 1998/02/20 01:00:16 manchek Exp $";

/*	img_new()
*
*	Create a new image with specified size
*/

struct img *
img_new(wd, ht)
	int wd;
	int ht;
{
	struct img *ip;

	if ((ip = TALLOC(1, struct img, 0))) {
		ip->wd = wd;
		ip->ht = ht;
		if ((ip->data = TALLOC(wd * ht, struct pixel, 0))) {
			BZERO(ip->data, wd * ht * sizeof(struct pixel));

		} else {
			MYFREE(ip);
			ip = 0;
		}
	}
	return ip;
}


int
img_free(imp)
	struct img *imp;
{
	if (imp->data)
		MYFREE(imp->data);
	MYFREE(imp);
	return 0;
}


/*	img_plot()
*
*	Plot a pixel in an image
*/

int
img_plot(imp, px, x, y)
	struct img *imp;
	struct pixel px;
	int x, y;
{

	if (x < 0 || x >= imp->wd || y < 0 || y >= imp->ht) {
/*
		fprintf(stderr, "img_plot() out of range %d, %d\n", x, y);
		abort();
*/
		return 0;
	}
	imp->data[x + imp->wd * y] = px;

	return 0;
}


/*	img_drawline()
*
*	Draw a line in an image using Bresenham's algorithm.
*/

int
img_drawline(imp, px, x1, y1, x2, y2)
	struct img *imp;
	struct pixel px;
	int x1, y1;
	int x2, y2;
{
	int big, small;		/* dimensions of rectangle to draw across */
	int dx, dy;			/* delta x, y */
	int xsgn, ysgn;		/* x, y direction */
	int x, y;			/* position accumulator */
	int a;				/* accum for interpolating */
	int i;

	if ((dx = x2 - x1) < 0) {
		dx = -dx;
		xsgn = -1;
	} else
		xsgn = 1;
	if ((dy = y2 - y1) < 0) {
		dy = -dy;
		ysgn = -1;
	} else
		ysgn = 1;

	big = max(dx, dy);
	small = min(dx, dy);

	x = x1;
	y = y1;

	a = big / 2;	/* but with symmetrical lines */
	for (i = 0; i <= big; i++) {
		img_plot(imp, px, x, y);
		if ((a += small) > big) {
			a -= big;
			if (dx < dy)
				x += xsgn;
			else
				y += ysgn;
		}
		if (dy > dx)
			y += ysgn;
		else
			x += xsgn;
	}
	return 0;
}


unsigned char cdata[96][30] = {
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x90, 0x00, 0x90, 0x00, 0x90, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x48, 0x00, 0x48, 0x00,
  0xfc, 0x00, 0x48, 0x00, 0x48, 0x00, 0xfc, 0x00, 0x48, 0x00,
  0x48, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0xf8, 0x00,
  0x24, 0x00, 0x24, 0x00, 0xf8, 0x00, 0x20, 0x01, 0x20, 0x01,
  0xf8, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x08, 0x01, 0x94, 0x00, 0x94, 0x00,
  0x48, 0x00, 0x20, 0x00, 0x20, 0x00, 0x90, 0x00, 0x48, 0x01,
  0x48, 0x01, 0x84, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x24, 0x00, 0x24, 0x00,
  0x24, 0x00, 0x18, 0x00, 0x24, 0x00, 0x44, 0x01, 0x84, 0x00,
  0x44, 0x01, 0x38, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xe0, 0x00, 0x60, 0x00, 0x10, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x88, 0x00, 0x50, 0x00, 0xfc, 0x01, 0x50, 0x00, 0x88, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0xfc, 0x01, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x70, 0x00, 0x30, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x20, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x80, 0x00, 0x80, 0x00,
  0x40, 0x00, 0x20, 0x00, 0x20, 0x00, 0x10, 0x00, 0x08, 0x00,
  0x08, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x30, 0x00, 0x48, 0x00, 0x84, 0x00,
  0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00,
  0x48, 0x00, 0x30, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x30, 0x00, 0x28, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x00, 0x01, 0x80, 0x00, 0x40, 0x00, 0x30, 0x00, 0x08, 0x00,
  0x04, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x00, 0x01, 0x80, 0x00,
  0x40, 0x00, 0xe0, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0xc0, 0x00, 0xa0, 0x00,
  0x90, 0x00, 0x88, 0x00, 0x84, 0x00, 0xfc, 0x01, 0x80, 0x00,
  0x80, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x04, 0x00, 0x04, 0x00,
  0xf4, 0x00, 0x0c, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf0, 0x00, 0x08, 0x00, 0x04, 0x00,
  0x04, 0x00, 0xf4, 0x00, 0x0c, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x80, 0x00, 0x40, 0x00, 0x20, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x84, 0x01, 0x78, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x80, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x20, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x20, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x20, 0x00, 0x70, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x70, 0x00, 0x30, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x40, 0x00, 0x20, 0x00,
  0x10, 0x00, 0x08, 0x00, 0x08, 0x00, 0x10, 0x00, 0x20, 0x00,
  0x40, 0x00, 0x80, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x08, 0x00, 0x10, 0x00, 0x20, 0x00,
  0x40, 0x00, 0x80, 0x00, 0x80, 0x00, 0x40, 0x00, 0x20, 0x00,
  0x10, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf0, 0x00, 0x08, 0x01, 0x08, 0x01,
  0x00, 0x01, 0x80, 0x00, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00,
  0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0xe4, 0x01, 0x14, 0x01, 0x94, 0x01, 0x64, 0x01, 0x04, 0x00,
  0x04, 0x00, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x50, 0x00, 0x88, 0x00,
  0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0xfc, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x08, 0x01, 0x08, 0x01,
  0x08, 0x01, 0xf8, 0x00, 0x08, 0x01, 0x08, 0x01, 0x08, 0x01,
  0x08, 0x01, 0xfc, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x00,
  0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x08, 0x01, 0x08, 0x01,
  0x08, 0x01, 0x08, 0x01, 0x08, 0x01, 0x08, 0x01, 0x08, 0x01,
  0x08, 0x01, 0xfc, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x08, 0x00, 0x08, 0x00,
  0x08, 0x00, 0x78, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00,
  0x08, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x08, 0x00, 0x08, 0x00,
  0x08, 0x00, 0x78, 0x00, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00,
  0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x00,
  0x04, 0x00, 0x04, 0x00, 0xc4, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xfc, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xe0, 0x01, 0x80, 0x00, 0x80, 0x00,
  0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00,
  0x84, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x84, 0x00, 0x44, 0x00,
  0x24, 0x00, 0x1c, 0x00, 0x14, 0x00, 0x24, 0x00, 0x44, 0x00,
  0x84, 0x00, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x8c, 0x01,
  0x54, 0x01, 0x54, 0x01, 0x24, 0x01, 0x24, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x0c, 0x01,
  0x14, 0x01, 0x24, 0x01, 0x44, 0x01, 0x84, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xfc, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x24, 0x01,
  0x44, 0x01, 0xf8, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xfc, 0x00, 0x24, 0x00, 0x44, 0x00, 0x84, 0x00,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x00,
  0x04, 0x00, 0xf8, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x88, 0x00, 0x88, 0x00, 0x88, 0x00, 0x50, 0x00, 0x50, 0x00,
  0x50, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x24, 0x01, 0x24, 0x01, 0x24, 0x01, 0x24, 0x01,
  0x54, 0x01, 0x88, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x88, 0x00,
  0x50, 0x00, 0x20, 0x00, 0x20, 0x00, 0x50, 0x00, 0x88, 0x00,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x04, 0x01, 0x88, 0x00,
  0x50, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xfc, 0x01, 0x00, 0x01, 0x80, 0x00,
  0x40, 0x00, 0x20, 0x00, 0x10, 0x00, 0x08, 0x00, 0x04, 0x00,
  0x04, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xf0, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x10, 0x00, 0xf0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x08, 0x00, 0x08, 0x00,
  0x10, 0x00, 0x20, 0x00, 0x20, 0x00, 0x40, 0x00, 0x80, 0x00,
  0x80, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x40, 0x00, 0x40, 0x00,
  0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00, 0x40, 0x00,
  0x40, 0x00, 0x78, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x50, 0x00, 0x88, 0x00,
  0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0xfe, 0x01, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x30, 0x00, 0x40, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf8, 0x00, 0x00, 0x01, 0x00, 0x01, 0xf8, 0x01, 0x04, 0x01,
  0x84, 0x01, 0x78, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0xf4, 0x00, 0x0c, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x0c, 0x01, 0xf4, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf8, 0x00, 0x04, 0x01, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x00, 0xf8, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01,
  0x78, 0x01, 0x84, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x84, 0x01, 0x78, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf8, 0x00, 0x04, 0x01, 0x04, 0x01, 0xfc, 0x01, 0x04, 0x00,
  0x04, 0x00, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xe0, 0x00, 0x10, 0x01, 0x10, 0x00,
  0x10, 0x00, 0x78, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x10, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x78, 0x01, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x78, 0x00,
  0x04, 0x00, 0xf8, 0x00, 0x04, 0x01, 0x04, 0x01, 0xf8, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0xf4, 0x00, 0x0c, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00,
  0x30, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00,
  0xe0, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x80, 0x00,
  0x80, 0x00, 0x80, 0x00, 0x80, 0x00, 0x84, 0x00, 0x78, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00,
  0x04, 0x01, 0xc4, 0x00, 0x34, 0x00, 0x0c, 0x00, 0x34, 0x00,
  0xc4, 0x00, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x30, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xdc, 0x00, 0x24, 0x01, 0x24, 0x01, 0x24, 0x01, 0x24, 0x01,
  0x24, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf4, 0x00, 0x0c, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf8, 0x00, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf4, 0x00, 0x0c, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x0c, 0x01, 0xf4, 0x00, 0x04, 0x00, 0x04, 0x00, 0x04, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x78, 0x01, 0x84, 0x01, 0x04, 0x01, 0x04, 0x01, 0x04, 0x01,
  0x84, 0x01, 0x78, 0x01, 0x00, 0x01, 0x00, 0x01, 0x00, 0x01 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xe4, 0x00, 0x18, 0x01, 0x08, 0x00, 0x08, 0x00, 0x08, 0x00,
  0x08, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xf8, 0x00, 0x04, 0x01, 0x04, 0x00, 0xf8, 0x00, 0x00, 0x01,
  0x04, 0x01, 0xf8, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x78, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00, 0x10, 0x00,
  0x10, 0x01, 0xe0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00,
  0x84, 0x00, 0x78, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x04, 0x01, 0x04, 0x01, 0x88, 0x00, 0x88, 0x00, 0x50, 0x00,
  0x50, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x04, 0x01, 0x04, 0x01, 0x24, 0x01, 0x24, 0x01, 0x24, 0x01,
  0x54, 0x01, 0x88, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x04, 0x01, 0x88, 0x00, 0x50, 0x00, 0x20, 0x00, 0x50, 0x00,
  0x88, 0x00, 0x04, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00, 0x84, 0x00,
  0xc4, 0x00, 0xb8, 0x00, 0x80, 0x00, 0x80, 0x00, 0x78, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0xfc, 0x01, 0x80, 0x00, 0x40, 0x00, 0x20, 0x00, 0x10, 0x00,
  0x08, 0x00, 0xfc, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0xc0, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x10, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0xc0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x18, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x40, 0x00, 0x20, 0x00, 0x20, 0x00, 0x20, 0x00,
  0x20, 0x00, 0x18, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x18, 0x01, 0x24, 0x01, 0xc4, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 },
{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }
};


/*	img_drawchar()
*
*	Draw a character in an image
*/

int
img_drawchar(imp, cp1, cp2, x, y, c)
	struct img *imp;
	struct pixel *cp1;
	struct pixel *cp2;
	int x;
	int y;
	char c;
{
	int x1, y1;
	unsigned char *p;
	struct pixel *cp;

	c &= 127;
	if (c < 32)
		c = ' ';
	p = cdata[c - ' '];
	for (y1 = 0; y1 < 15; y1++)
		for (x1 = 0; x1 < 11; x1++) {
			if ((cp = (p[y1 * 2 + x1 / 8] & (1 << (x1 & 7))) ? cp1 : cp2))
				img_plot(imp, *cp, x + x1, y - FONT_ASCENT + y1);
		}
	return 0;
}


/*	img_drawstring()
*
*	Draw a string in an image
*/

int
img_drawstring(imp, cp1, cp2, x, y, s, l)
	struct img *imp;
	struct pixel *cp1;
	struct pixel *cp2;
	int x;
	int y;
	char *s;
	int l;
{
/*
	fprintf(stderr, "drawstring(%s)\n", s);
*/
	while (l-- > 0 && *s) {
		img_drawchar(imp, cp1, cp2, x, y, *s);
		x += 11;
		s++;
	}
	return 0;
}


/*	img_textwidth()
*
*	Determine pixel width of text in image
*/

int
img_textwidth(imp, s, len)
	struct img *imp;
	char *s;
	int len;
{
	return len * 11;
}


int
img_fontattr(asc, des)
	int *asc;
	int *des;
{
	*asc = FONT_ASCENT;
	*des = FONT_DESCENT;
	return 0;
}


int
img_blend(front, back, dest)
	struct img *front;
	struct img *back;
	struct img *dest;
{
	int i;
	struct pixel fpx, bpx, dpx;

	for (i = dest->wd * dest->ht; i-- > 0; ) {
		fpx = front->data[i];
		bpx = back->data[i];
		dpx.r = ((int)fpx.r * fpx.a + (int)bpx.r * (255 - fpx.a)) / 255;
		dpx.g = ((int)fpx.g * fpx.a + (int)bpx.g * (255 - fpx.a)) / 255;
		dpx.b = ((int)fpx.b * fpx.a + (int)bpx.b * (255 - fpx.a)) / 255;
		dpx.a = (int)fpx.a + (int)bpx.a - ((int)fpx.a * bpx.a) / 255;
		dest->data[i] = dpx;
	}
	return 0;
}


int
img_write(name, ff, imp)
	char *name;
	FILE *ff;
	struct img *imp;
{
	int i;
	char *p;

	if (!(p = TALLOC(imp->wd * imp->ht * 3, char, 0)))
		return 1;

	for (i = imp->wd * imp->ht; i-- > 0; ) {
		p[i * 3] = imp->data[i].r;
		p[i * 3 + 1] = imp->data[i].g;
		p[i * 3 + 2] = imp->data[i].b;
	}
	writeppm(name, ff, imp->wd, imp->ht, p);
	MYFREE(p);
	return 0;
}


struct img *
img_read(name, ff)
	char *name;
	FILE *ff;
{
	struct img *imp;
	int wd;
	int ht;
	char *data;
	struct pixel px;
	int i;

	if (readppm(name, ff, &wd, &ht, &data))
		return (struct img *)0;
	if (!(imp = img_new(wd, ht))) {
		MYFREE(data);
		return (struct img *)0;
	}
	for (i = wd * ht; i-- > 0; ) {
		px.r = data[i * 3];
		px.g = data[i * 3 + 1];
		px.b = data[i * 3 + 2];
		px.a = 255;
		imp->data[i] = px;
	}
	MYFREE(data);
	return imp;
}


/*	xdig()
*
*	Return value of hexadecimal character
*/

static int
xdig(c)
	char c;
{
	return (c >= '0' && c <= '9') ? c - '0'
			: ((c >= 'A' && c <= 'F') ? c - 'A' + 10
			: ((c >= 'a' && c <= 'f') ? c - 'a' + 10
			: 0));
}


/*	parsecolor()
*
*	Parse a color name (in format #xxx or #xxxxxx)
*/

int
parsecolor(name, col)
	char *name;
	struct pixel *col;
{
	if (name[0] != '#') {
		fprintf(stderr, "can't deal with color names: %s\n", name);
		return 1;
	}
	name++;
	switch (strlen(name)) {
		case 3:
			col->r = xdig(name[0]) * 16;
			col->g = xdig(name[1]) * 16;
			col->b = xdig(name[2]) * 16;
			col->a = 255;
			break;

		case 4:
			col->r = xdig(name[0]) * 16;
			col->g = xdig(name[1]) * 16;
			col->b = xdig(name[2]) * 16;
			col->a = xdig(name[3]) * 16;
			break;

		case 6:
			col->r = xdig(name[0]) * 16 + xdig(name[1]);
			col->g = xdig(name[2]) * 16 + xdig(name[3]);
			col->b = xdig(name[4]) * 16 + xdig(name[5]);
			col->a = 255;
			break;

		case 8:
			col->r = xdig(name[0]) * 16 + xdig(name[1]);
			col->g = xdig(name[2]) * 16 + xdig(name[3]);
			col->b = xdig(name[4]) * 16 + xdig(name[5]);
			col->a = xdig(name[6]) * 16 + xdig(name[7]);
			break;

		default:
			fprintf(stderr, "can't convert color: %s\n", --name);
			return 1;
			break;
	}
	return 0;
}


int
img_shadow(imp, dx, dy, scolor)
	struct img *imp;
	double dx;
	double dy;
	struct pixel scolor;
{
	struct img *sha;
	int idx, idy, x, y;
	struct pixel px;

	idx = dx;
	idy = dy;
	if ((sha = img_new(imp->wd, imp->ht))) {
		for (y = 0; y < imp->ht; y++)
			for (x = 0; x < imp->wd; x++) {
				px = scolor;
				px.a = ((int)px.a * imp->data[x + imp->wd * y].a) / 255;
				img_plot(sha, px, x + idx, y + idy);
			}
		img_blend(imp, sha, imp);
		img_free(sha);

	} else
		return 1;
	return 0;
}


