
/*
*	$Id: logfile.c,v 1.3 1998/09/09 01:17:55 manchek Exp $
*
*	Read and repair temperature log files.
*	10 nov 1996  Robert Manchek
*/

/*
*  Copyright (C) 1995, 1996 by Robert Manchek.  (manchek@spiderplant.com)
*
*  Permission to use, copy, modify, and distribute this software
*  and its documentation for any purpose and without fee is hereby
*  granted, provided that this copyright notice appear in all
*  copies and that both that copyright notice and this permission
*  notice appear in supporting documentation.
*
*  This software is provided "as is" without express or implied
*  warranty.
*/

/*
*	$Log: logfile.c,v $
*	Revision 1.3  1998/09/09 01:17:55  manchek
*	y2k fix - if yr < 100(70), add 1900(2000)
*
*	Revision 1.2  1996/11/12 06:13:01  manchek
*	code cleanup
*
*	Revision 1.1  1996/11/12 05:19:19  manchek
*	Initial revision
*
*/


#include <stdio.h>
#include <imalloc.h>
#include "logfile.h"

static char rcsid[] = "$Id: logfile.c,v 1.3 1998/09/09 01:17:55 manchek Exp $";
static char daysinmonth[] = { 0,
	31, 28, 31, 30,
	31, 30, 31, 31,
	30, 31, 30, 31
};


/*	tonext()
*
*	update time to time of next expected sample.
*/

int
tonext(t)
	struct sample *t;
{
	/* snap time to interval */
	t->mn = ((t->mn + SAMPLEINTERVAL/2) / SAMPLEINTERVAL) * SAMPLEINTERVAL;
	t->mn += SAMPLEINTERVAL;

	while (t->mn > 59) {
		t->mn -= 60;
		if (++t->hr > 23) {
			t->hr = 0;
			if (++t->da > daysinmonth[t->mo]) {
				t->da = 1;
				if (++t->mo > 12) {
					t->mo = 1;
					t->yr++;
				}
			}
		}
	}
	return 0;
}


int
tdiff(t1, t2)
	struct sample *t1;
	struct sample *t2;
{
	int m1, m2;

	m1 = (((t1->yr * 12 + t1->mo) * 31 + t1->da) * 24 + t1->hr) * 60 + t1->mn;
	m2 = (((t2->yr * 12 + t2->mo) * 31 + t2->da) * 24 + t2->hr) * 60 + t2->mn;
	return m1 - m2;
}


/*	readlogfile()
*
*	Read the last (count) samples from a logfile into an array of
*	samples.  Fill in any missing ones with placeholders.
*
*	Returns numbers of samples found, else -1 for error.
*/

int
readlogfile(ff, data, count)
	FILE *ff;
	struct sample *data;
	int count;
{
	struct sample this;
	struct sample last;
	int once = 1;
	int i;
	int got;
	int wp;
	struct sample *sambuf;
	char buf[80];

	if (!(sambuf = TALLOC(count, struct sample, 0)))
		return -1;
	got = 0;
	wp = 0;
	while (fgets(buf, sizeof(buf)-1, ff)) {
		i = sscanf(buf, " %d %d %d %d %d %f",
				&this.yr, &this.mo, &this.da, &this.hr, &this.mn, &this.val);
		if (i > 4) {
			if (this.yr < 100) {
				if (this.yr < 70)
					this.yr += 2000;
				else
					this.yr += 1900;
			}
			if (i < 6)
				this.val = NOREADING;
			if (once) {
				once = 0;
			} else {
				tonext(&last);
				while (tdiff(&last, &this) < -SAMPLEINTERVAL/2) {
/*
					printf("%02d %02d %02d %02d %02d ?\n",
						last.yr, last.mo, last.da, last.hr, last.mn);
*/
					last.val = NOREADING;
					sambuf[wp] = last;
					got++;
					if (++wp >= count)
						wp = 0;
					tonext(&last);
				}
			}
			last = this;
/*
			printf("%02d %02d %02d %02d %02d %.1f\n",
				this.yr, this.mo, this.da, this.hr, this.mn, this.val);
*/
			sambuf[wp] = this;
			got++;
			if (++wp >= count)
				wp = 0;
		}
	}
	if (got > count)
		got = count;
	wp -= got;
	if (wp < 0)
		wp += count;

	for (i = 0; i < got; i++) {
/*
		printf("%02d %02d %02d %02d %02d %.1f\n",
			sambuf[wp].yr, sambuf[wp].mo, sambuf[wp].da, sambuf[wp].hr, sambuf[wp].mn,
			sambuf[wp].val);
*/
		data[i] = sambuf[wp];
		if (++wp >= count)
			wp = 0;
	}
	MYFREE(sambuf);

	return got;
}


