
/*
*	$Id: ppm.c,v 1.2 1996/11/12 06:08:19 manchek Exp $
*
*	Minimalist PPM read/write routines
*	06 Feb 1996 Robert Manchek
*/

/*
*  Copyright (C) 1995, 1996 by Robert Manchek.  (manchek@spiderplant.com)
*
*  Permission to use, copy, modify, and distribute this software
*  and its documentation for any purpose and without fee is hereby
*  granted, provided that this copyright notice appear in all
*  copies and that both that copyright notice and this permission
*  notice appear in supporting documentation.
*
*  This software is provided "as is" without express or implied
*  warranty.
*/

/*
*	$Log: ppm.c,v $
*	Revision 1.2  1996/11/12 06:08:19  manchek
*	code cleanup
*
*	Revision 1.1  1996/11/12 05:19:19  manchek
*	Initial revision
*
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <imalloc.h>

static char rcsid[] = "$Id: ppm.c,v 1.2 1996/11/12 06:08:19 manchek Exp $";


static int
skipws(ff)
	FILE *ff;
{
	int c;

	c = getc(ff);
	while (isspace(c) || c == '#') {
		if (c == '#') {
			while (c = getc(ff), c != EOF && c != '\n') ;
			if (c != EOF)
				c = getc(ff);
		} else
			while (c = getc(ff), c != EOF && isspace(c)) ;
	}
	ungetc(c, ff);
	return 0;
}


static int
getint(ff, np)
	FILE *ff;
	int *np;
{
	int c;
	int n = 0;

	skipws(ff);
	c = getc(ff);
	if (!isdigit(c))
		return 1;
	do {
		n = n * 10 + c - '0';
	} while (c = getc(ff), isdigit(c));
	ungetc(c, ff);
	*np = n;
	return 0;
}


int
readppm(name, ff, wdp, htp, datap)
	char *name;
	FILE *ff;
	int *wdp;
	int *htp;
	char **datap;
{
	int c;
	int pixmax;

	if (getc(ff) != 'P' || getc(ff) != '6') {
		fprintf(stderr, "%s: not P6 format\n", name);
		return 1;
	}
	if (getint(ff, wdp) || getint(ff, htp) || getint(ff, &pixmax)) {
		fprintf(stderr, "%s: bad P6 header\n", name);
		return 1;
	}
	while (c = getc(ff), c != '\n') ;

	*datap = TALLOC(*wdp * *htp * 3, char, "ppm");
	if (fread(*datap, 3, *wdp * *htp, ff) != *wdp * *htp) {
		fprintf(stderr, "%s: short read of P6 file\n", name);
		MYFREE(*datap);
		return 1;
	}

	return 0;
}


int
writeppm(name, ff, wd, ht, data)
	char *name;
	FILE *ff;
	int wd;
	int ht;
	char *data;
{
	fprintf(ff, "P6\n# comment\n%d %d\n%d\n", wd, ht, 255);
	fwrite(data, 3, wd * ht, ff);
	return 0;
}


