
/*
*	$Id: thermplot.c,v 1.9 1998/09/09 01:38:44 manchek Exp $
*
*	Combine a backrgound image with temperature plots and spit
*	out a PPM file.
*	11 Nov 1996  Robert Manchek
*/

/*
*  Copyright (C) 1995, 1996 by Robert Manchek.  (manchek@spiderplant.com)
*
*  Permission to use, copy, modify, and distribute this software
*  and its documentation for any purpose and without fee is hereby
*  granted, provided that this copyright notice appear in all
*  copies and that both that copyright notice and this permission
*  notice appear in supporting documentation.
*
*  This software is provided "as is" without express or implied
*  warranty.
*/

/*
*	$Log: thermplot.c,v $
*	Revision 1.9  1998/09/09 01:38:44  manchek
*	compile-time switch to show day name along with date in graph
*
*	Revision 1.8  1998/02/20 01:00:50  manchek
*	added multiline capability for graph title string
*
*	Revision 1.7  1997/12/22 23:02:27  manchek
*	rounded up vmax by 1 to scale top of graph correctly
*
*	Revision 1.6  1996/11/25 03:15:01  manchek
*	typo in draw_graph, some sample sets may be shorter than others
*
*	Revision 1.5  1996/11/17 01:19:59  manchek
*	finished c or f conversion option.
*	added -j to specify label step
*
*	Revision 1.4  1996/11/14 04:57:11  manchek
*	added default graph color table
*
*	Revision 1.3  1996/11/14 04:40:17  manchek
*	graph grid goes behind plot if filled
*
*	Revision 1.2  1996/11/12 06:16:03  manchek
*	code cleanup
*
*	Revision 1.1  1996/11/12 05:19:19  manchek
*	Initial revision
*
*/


#ifndef	DOLOG
#define	DOLOG	0
#endif

#include <stdlib.h>
#include <stdio.h>
#ifdef	SYSVSTR
#include <string.h>
#define	CINDEX(s,c)	strchr(s,c)
#else
#include <strings.h>
#define	CINDEX(s,c)	index(s,c)
#endif
#include <ctype.h>
#include <math.h>
#include <sys/types.h>
#include <sys/time.h>
#if	DOLOG
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <syslog.h>
#endif
#include <imalloc.h>
#include "logfile.h"
#include "draw.h"

#ifndef	DEF_LABEL
#define	DEF_LABEL		'b'
#endif
#ifndef	DEF_TITLE
#define	DEF_TITLE		0
#endif
#ifndef	DEF_IMAGE_WIDTH
#define	DEF_IMAGE_WIDTH		200
#endif
#ifndef	DEF_IMAGE_HEIGHT
#define	DEF_IMAGE_HEIGHT	150
#endif
#ifndef	DEF_BGC
#define	DEF_BGC		"#0000"
#endif
#ifndef	DEF_SDC
#define	DEF_SDC		"#0008"
#endif
#ifndef	DEF_AXC
#define	DEF_AXC		"#00a0c0"
/*
#define	DEF_AXC		"#853e26"
*/
#endif

#define	BGCOLOR		0
#define	AXCOLOR		1
#define	SDCOLOR		2
#define	GRCOLOR		3

#ifndef	min
#define	min(a,b)	((a)<(b)?(a):(b))
#endif
#ifndef	max
#define	max(a,b)	((a)>(b)?(a):(b))
#endif

struct dataset {
	struct sample *samples;
	int count;

	char *file;
	struct pixel px;
};

struct plot {
	char *title;		/* title string */
	char label;			/* where to label degrees */
	char conv;			/* conversion to perform */
	int solid;			/* draw filled graph or lines */
	int step;			/* degree mark step */
};


/************
*  Globals  *
*           *
************/

static char rcsid[] = "$Id: thermplot.c,v 1.9 1998/09/09 01:38:44 manchek Exp $";

/* the default color list for -g */
static char *defcols[] = {
	"#ff0000",
	"#ff8000",
	"#ffff00",
	"#00ff00",
	"#0000ff",
	"#ff00ff",
	"#808080",
	"#ffffff",
};


int
istic(hr, mn)
	int hr, mn;
{
	if (hr >= 9 && hr < 15)
		return mn < 45 ? 1 : 0;
	if (hr < 3 || hr >= 21)
		return mn < 15 ? 1 : 0;
	return mn < 30 ? 1 : 0;
}


/*	draw_graph()
*
*	Draw the whole temperature graph with labels and stuff
*/

int
draw_graph(plat, imp, nsets, sets, pixs)
	struct plot *plat;			/* plot attributes */
	struct img *imp;			/* base image */
	int nsets;					/* number of data sets */
	struct dataset **sets;
	struct pixel pixs[4];
{
	int wd, ht;					/* of image */
	int oxo, oyo, owd, oht;		/* whole overlay left, top, width, height */
	int gxo, gyo, gwd, ght;		/* graph left, top, width, height */
	int n;
	int i;
	int x, y, y0;
	int vmin, vmax;
	int lhow;
	int ascent, descent;
	int s;
	struct sample sam;
	int afew;
	int tlines = 0;
	char buf[200];

	if (nsets < 1)
		return 0;

	wd = imp->wd;
	ht = imp->ht;
	oxo = 0;
	oyo = 0;
	owd = wd;
	oht = ht;
	img_fontattr(&ascent, &descent);

	img_drawline(imp, pixs[AXCOLOR], oxo, oyo, oxo + owd - 1, oyo);
	img_drawline(imp, pixs[AXCOLOR], oxo, oyo + oht - 1, oxo + owd - 1, oyo + oht - 1);
	img_drawline(imp, pixs[AXCOLOR], oxo, oyo + 1, oxo, oyo + oht - 2);
	img_drawline(imp, pixs[AXCOLOR], oxo + owd - 1, oyo + 1, oxo + owd - 1, oyo + oht - 2);

	lhow = ((plat->label == 'l' || plat->label == 'b') ? 1 : 0)
			+ ((plat->label == 'r' || plat->label == 'b') ? 2 : 0);

	gxo = img_textwidth(imp, "188", 3);
	switch (lhow) {
	case 1:
		gwd = owd - gxo;
		break;
	case 2:
		gwd = owd - gxo;
		gxo = 0;
		break;
	case 3:
		gwd = owd - 2 * gxo;
		break;
	case 0:
		gwd = owd;
		gxo = 0;
		break;
	}
	gxo += oxo;
	gyo = 3 + ascent / 2;
	if (plat->title && plat->title[0]) {
		char *p;

		tlines = 1;
		for (p = plat->title; p = CINDEX(p, '\n'); p++)
			tlines++;
	}
	ght = oht - (ascent + descent) / 2 - gyo - tlines * (ascent + descent) - 1;
	gyo += oyo;

	/*
	* get min and max temperatures for all data sets
	*/

	vmin = 200;
	vmax = -200;
	for (s = 0; s < nsets; s++) {
		n = min(gwd, sets[s]->count);
		for (i = 0; i < n; i++) {
			y = sets[s]->samples[sets[s]->count - n + i].val;
			if (y != NOREADING) {
				if (y < vmin)
					vmin = y;
				if (y > vmax)
					vmax = y;
			}
		}
	}
	vmax++;
	if (vmin > vmax)
		vmin = vmax = 0;
	if (vmin < 0)
		vmin = ((vmin - 9) / 10) * 10;
	else
		vmin = (vmin / 10) * 10;
	if (vmax < 0)
		vmax = (vmax / 10) * 10;
	else
		vmax = ((vmax + 9) / 10) * 10;
	if (vmin == vmax)
		vmax += 10;
/*
	fprintf(stderr, "%d to %d\n", vmin, vmax);
*/

	afew = 0;
	do {

	/*
	* draw day numbers and hour scale tic marks
	*/

		n = min(gwd, sets[0]->count);
		for (i = 0; i < n; i++) {
			sam = sets[0]->samples[sets[0]->count - n + i];
			if (sam.mn < 15) {
				if (sam.hr == 0) {
					x = gxo + gwd - n + i;
					img_drawline(imp, pixs[AXCOLOR], x, gyo,
							x, gyo + ascent + descent + 1);
				}
				if (sam.hr == 12) {
#ifdef DISPLAY_DAYNAME
					static char *daynames[] = {
						"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat",
					};
					struct tm tm, *ptm;
					time_t ti;
					tm.tm_sec = 0;
					tm.tm_min = sam.mn;
					tm.tm_hour = sam.hr;
					tm.tm_mday = sam.da;
					tm.tm_mon = sam.mo - 1;
					tm.tm_year = (sam.yr < 100 ? sam.yr : sam.yr - 1900);
					tm.tm_wday = 0;
					tm.tm_yday = 0;
					tm.tm_isdst = 0;
					ti = mktime(&tm);
					ptm = localtime(&ti);
					sprintf(buf, "%d/%d %s", ptm->tm_mon + 1, ptm->tm_mday,
						daynames[ptm->tm_wday]);
#else /*DISPLAY_DAYNAME*/
					sprintf(buf, "%d", sam.da);
#endif /*DISPLAY_DAYNAME*/
					x = gxo + gwd - n + i - img_textwidth(imp, buf, strlen(buf)) / 2;
					img_drawstring(imp, &pixs[AXCOLOR], (struct pixel *)0,
							x, gyo + ascent + 2,
							buf, strlen(buf));
				}
			}
			if (istic(sam.hr, sam.mn)) {
				x = gxo + gwd - n + i;
				img_drawline(imp, pixs[AXCOLOR], x, oyo + 1, x, oyo + 3);
			}
		}

	/*
	* draw temperature scale
	*/

		if (lhow)
			for (i = vmin; i <= vmax; i += plat->step) {
				sprintf(buf, "%d", i);
				y = gyo + ght - 1 - ((i - vmin) * (ght - 1)) / (vmax - vmin);
				img_drawline(imp, pixs[AXCOLOR], gxo, y, gxo + gwd - 1, y);
				if (lhow & 1) {
					img_drawstring(imp, &pixs[AXCOLOR], (struct pixel *)0,
							oxo + 2, y + ascent / 2,
							buf, strlen(buf));
				}
				if (lhow & 2) {
					img_drawstring(imp, &pixs[AXCOLOR], (struct pixel *)0,
							gxo + gwd + 2, y + ascent / 2,
							buf, strlen(buf));
				}
			}

		if (plat->solid && afew)
			break;
		afew++;

	/*
	* draw all temperature plots
	*/

		for (s = 0; s < nsets; s++) {
			n = min(gwd, sets[s]->count);
			y0 = -1;
			for (i = 0; i < n; i++) {
				x = gxo + gwd - n + i;
				sam = sets[s]->samples[sets[s]->count - n + i];
				if (sam.val != NOREADING) {
					y = gyo + ght - 1
							- ((sam.val - vmin) * (ght - 1)) / (vmax - vmin);
					if (plat->solid) {
						img_drawline(imp, sets[s]->px, x, y, x, ht - 1);

					} else {
						if (y0 == -1) {
							img_drawline(imp, sets[s]->px, x, y, x, y);

						} else {
							if (y < y0) {
								img_drawline(imp, sets[s]->px, x, y0, x, y);

							} else {
								img_drawline(imp, sets[s]->px, x - 1, y0, x - 1, y);
								img_drawline(imp, sets[s]->px, x - 1, y, x, y);
							}
						}
					}
					y0 = y;
				} else
					y0 = -1;
			}
		}
	} while (plat->solid);

	if (plat->title && plat->title[0]) {
		char *p, *q;

		p = plat->title;
		for (i = 1; i <= tlines; i++) {
			if (!(q = CINDEX(p, '\n')))
				q = p + strlen(p);
			img_drawstring(imp, &pixs[AXCOLOR], (struct pixel *)0,
				oxo + (owd - img_textwidth(imp, p, q - p)) / 2,
				oyo + oht - (tlines - i) * (ascent + descent) - descent - 2,
				p, q - p);
			p = q + 1;
		}
	}

	return 0;
}


#if	DOLOG
/*	iden()
*
*	syslog a line identifying the peer socket (like if we're started by inetd)
*/

iden(d)
	int d;
{
	struct sockaddr_in sin;
	int l = sizeof(sin);
	unsigned int a;
	struct hostent *he;
	char buf[500];

	if (getpeername(d, (struct sockaddr *)&sin, &l) == 0) {
		if (he = gethostbyaddr((char *)&sin.sin_addr, sizeof(sin.sin_addr), AF_INET)) {
			syslog(LOG_INFO, "thermgif: from %s", he->h_name);
		} else {
			a = ntohl(sin.sin_addr.s_addr);
			syslog(LOG_INFO, "thermgif: from %d.%d.%d.%d",
					0xff & (a >> 24),
					0xff & (a >> 16),
					0xff & (a >> 8),
					0xff & a);
		}
	} else {
		syslog(LOG_INFO, "thermgif: getpeername: %m");
	}

	return 0;
}
#endif


int
main(argc, argv)
	int argc;
	char **argv;
{
	struct img *im1 = 0;
	struct img *im2 = 0;
#ifdef	PPMTOGIF
	int pfd[2];
#ifdef	PPMQUANT
	int p2fd[2];
#endif
#endif
	int i, j;
	int err = 0;
	char *p;
	int wd = -1;
	int ht = -1;
	char *bgimfile = 0;
	FILE *ff;
	struct pixel pixs[4];
	struct dataset **sets = 0;
	int msets = 0;
	int nsets = 0;
	struct plot plat;
	struct pixel *defpix;
	int ndefpix = sizeof(defcols)/sizeof(defcols[0]);
	int grcolset = 0;

	defpix = TALLOC(ndefpix, struct pixel, 0);
	for (i = 0; i < ndefpix; i++)
		parsecolor(defcols[i], &defpix[i]);

	plat.title = DEF_TITLE;
	plat.label = DEF_LABEL;
	plat.conv = 'n';
	plat.solid = 0;
	plat.step = 10;

	parsecolor(DEF_BGC, &pixs[BGCOLOR]);
	parsecolor(DEF_AXC, &pixs[AXCOLOR]);
	parsecolor(DEF_SDC, &pixs[SDCOLOR]);

	for (i = j = 1; i < argc; i++) {
		if (argv[i][0] == '-' && argv[i][1]) {
			for (p = argv[i] + 1; *p; p++) {
				switch (*p) {

				case 'w':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -w\n");
						err++;
					} else
						wd = atoi(argv[i]);
					break;

				case 'h':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -h\n");
						err++;
					} else
						ht = atoi(argv[i]);
					break;

				case 't':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -t\n");
						err++;
					} else
						plat.title = argv[i];
					break;

				case 'l':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -l\n");
						err++;
					} else
						plat.label = argv[i][0];
					break;

				case 's':
					plat.solid = 1;
					break;

				case 'j':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -j\n");
						err++;
					} else {
						plat.step = atoi(argv[i]);
						if (plat.step < 1)
							plat.step = 1;
					}
					break;

				case 'c':
				case 'f':
					plat.conv = *p;
					break;

				case 'b':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -b\n");
						err++;
					} else
						parsecolor(argv[i], &pixs[BGCOLOR]);
					break;

				case 'g':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -g\n");
						err++;
					} else {
						parsecolor(argv[i], &pixs[GRCOLOR]);
						grcolset = 1;
					}
					break;

				case 'a':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -a\n");
						err++;
					} else
						parsecolor(argv[i], &pixs[AXCOLOR]);
					break;

				case 'd':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -d\n");
						err++;
					} else
						parsecolor(argv[i], &pixs[SDCOLOR]);
					break;

				case 'i':
					if (++i == argc) {
						fprintf(stderr, "missing parameter for -h\n");
						err++;
					} else
						bgimfile = argv[i];
					break;

				default:
					fprintf(stderr, "unknown switch: %c\n", *p);
					err++;
					break;
				}
			}

		} else {
/*
			argv[j++] = argv[i];
*/
			if (nsets == msets) {
				msets = (msets * 8) / 5 + 1;
				if (sets)
					sets = TREALLOC(sets, msets, struct dataset *);
				else
					sets = TALLOC(msets, struct dataset *, 0);
			}
			sets[nsets] = TALLOC(1, struct dataset, 0);
			sets[nsets]->file = argv[i];
			sets[nsets]->px = grcolset ?
					pixs[GRCOLOR] : defpix[nsets % ndefpix];
			nsets++;
		}
	}
/*
	argc = j;
*/

	if (err || nsets < 1) {
		fprintf(stderr,
"usage: thermplot [ options ] logfile/- ...\n\
\n\
options: [-w wd] graph width\n\
         [-h ht] graph height\n\
         [-b #xxx[x]] background color, opacity\n\
         [-a #xxx[x]] axis and label color\n\
         [-g #xxx[x]] graph color\n\
         [-d #xxx[x]] drop shadow color\n\
         [-i name/-] background image\n\
         [-l lrbn] temperature labels on left, right, both, none\n\
         [-t title] title plot\n\
         [-s] fill area below graph\n\
         [-j step] set increment in degrees between graph labels\n\
         [-c] convert to celsius\n\
         [-f] convert to fahrenheit\n\
\n\
colors can be specified as 4 bits per channel (#xxx or #xxxx)\n\
or 8 bits (#xxxxxx or #xxxxxxxx)\n\
\n\
use -g more than once to make each plot a different color\n\
don't use -g to get different colors automatically\n\
\n\
input and output images are P6 PPM files\n\
\n\
");
		exit(1);
	}

#if	DOLOG
	openlog("thermgif", LOG_PID, LOG_LOCAL0);
	iden(0);
#endif

	if (bgimfile) {
		if (!strcmp(bgimfile, "-"))
			ff = stdin;
		else if (!(ff = fopen(bgimfile, "r"))) {
			fprintf(stderr, "%s: can't read\n", bgimfile);
			exit(1);
		}
		im1 = img_read(bgimfile, ff);
		if (ff != stdin)
			fclose(ff);
/*
		fprintf(stderr, "read %d x %d base image\n", im1->wd, im1->ht);
*/
		if (im1) {
			if (wd == -1)
				wd = im1->wd;
			if (ht == -1)
				ht = im1->ht;
		}
	}

	if (wd == -1)
		wd = DEF_IMAGE_WIDTH;
	if (ht == -1)
		ht = DEF_IMAGE_HEIGHT;
	im2 = img_new(wd, ht);
	for (i = 0; i < ht; i++)
		img_drawline(im2, pixs[BGCOLOR], 0, i, wd - 1, i);

#if 0
	nsets = argc - 1;
	sets = TALLOC(nsets, struct dataset *, 0);
	for (i = 1; i < argc; i++) {
		sets[i - 1] = TALLOC(1, struct dataset, 0);
		sets[i - 1]->file = argv[i];
		sets[i - 1]->px = pixs[GRCOLOR];
	}
#endif

	for (i = 0; i < nsets; i++) {
		sets[i]->count = wd;
		sets[i]->samples = TALLOC(wd, struct sample, 0);
		if (!strcmp(sets[i]->file, "-"))
			ff = stdin;
		else if (!(ff = fopen(sets[i]->file, "r"))) {
			fprintf(stderr, "%s: can't read\n", sets[i]->file);
			exit(1);
		}
		sets[i]->count = readlogfile(ff, sets[i]->samples, sets[i]->count);
		if (sets[i]->count == -1)
			sets[i]->count = 0;
		if (ff != stdin)
			fclose(ff);
		if (plat.conv == 'f')
			for (j = sets[i]->count; j-- > 0; )
				if (sets[i]->samples[j].val != NOREADING)
					sets[i]->samples[j].val =
							(sets[i]->samples[j].val * 9) / 5 + 32;
		if (plat.conv == 'c')
			for (j = sets[i]->count; j-- > 0; )
				if (sets[i]->samples[j].val != NOREADING)
					sets[i]->samples[j].val =
							((sets[i]->samples[j].val - 32) * 5) / 9;
	}

	draw_graph(&plat, im2, nsets, sets, pixs);
	if (im1) {
		img_shadow(im2, 1.0, 1.0, pixs[SDCOLOR]);
		img_blend(im2, im1, im1);
	} else
		im1 = im2;

#ifdef	PPMTOGIF
	pipe(pfd);

	if (!fork()) {
		close(2);
		close(pfd[1]);
		dup2(pfd[0], 0);
		execlp(PPMTOGIF, PPMTOGIF, (char *)0);
/*
		execlp(PPMTOGIF, PPMTOGIF, "-interlace", (char *)0);
*/
		perror(PPMTOGIF);
		_exit(1);
	}

	close(pfd[0]);
#ifdef	PPMQUANT
	pipe(p2fd);
	if (!fork()) {
		close(2);
		close(p2fd[1]);
		dup2(p2fd[0], 0);
		dup2(pfd[1], 1);
		execlp(PPMQUANT, PPMQUANT, "128", (char *)0);
		_exit(1);

	}
	close(p2fd[0]);
	dup2(p2fd[1], 1);
	close(pfd[1]);

#else
	dup2(pfd[1], 1);
#endif

#endif

	img_write("", stdout, im1);

#ifdef	PPMTOGIF
	fclose(stdout);
	wait((int *)0);
#ifdef	PPMQUANT
	wait((int *)0);
#endif
#endif

/*
	i_dump(1);
*/
	exit(0);
}


