
/*
*	$Id: xtherm.c,v 1.8 1998/09/09 01:25:56 manchek Exp $
*	xtherm.c
*
*	Display a temperature graph in an X-window, updated every 15 minutes.
*	Tails a log file created by reading a thermometer periodically.
*
*	21 mar 1995  Robert Manchek  manchek@cs.utk.edu
*
*	$Log: xtherm.c,v $
*	Revision 1.8  1998/09/09 01:25:56  manchek
*	compile-time switch to show day name along with date in graph
*
*	Revision 1.7  1997/12/22 23:00:03  manchek
*	rounded up vmax by 1 to scale top of graph correctly
*
*	Revision 1.6  1997/05/13 23:20:30  manchek
*	added -plot option to set graph colors
*
*	Revision 1.5  1996/11/25 00:34:05  manchek
*	typo in draw_graph, some sample sets may be shorter than others
*
*	Revision 1.4  1996/11/21 00:26:39  manchek
*	added -help and cleaned up usage string
*
*	Revision 1.3  1996/11/17 04:51:01  manchek
*	cleaned up the code a bit
*
*	Revision 1.2  1996/11/17 03:17:16  manchek
*	converted to multiple traces and new logfile reader
*
*	Revision 1.1  1996/10/29 03:28:21  manchek
*	Initial revision
*
*
*	17 mar 1996 switch to floating-point temperatures
*/

/*
*  Copyright (C) 1995 by Robert Manchek.  manchek@CS.UTK.EDU.
*
*  Permission to use, copy, modify, and distribute this software
*  and its documentation for any purpose and without fee is hereby
*  granted, provided that this copyright notice appear in all
*  copies and that both that copyright notice and this permission
*  notice appear in supporting documentation.
*
*  This software is provided "as is" without express or implied
*  warranty.
*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <imalloc.h>
#include "logfile.h"

#ifndef	min
#define	min(a,b)	((a)<(b)?(a):(b))
#endif
#ifndef	max
#define	max(a,b)	((a)>(b)?(a):(b))
#endif

#ifndef	DEF_LABEL
#define	DEF_LABEL		"b"
#endif
#ifndef	DEF_TITLE
#define	DEF_TITLE		0
#endif

#define	BGCOLOR		0
#define	AXCOLOR		1
#define	SDCOLOR		2
#define	GRCOLOR		3

#ifndef	LOGFILE
#define	LOGFILE	0
#endif

struct dataset {
	struct sample *samples;
	int count;

	char *file;
	GC gc;
};

struct plot {
	char *title;		/* title string */
	char label;			/* where to label degrees */
	char conv;			/* conversion to perform */
	int solid;			/* draw filled graph or lines */
	int step;			/* degree mark step */
};


#ifdef _PT_
#undef _PT_
#endif
#ifdef __STDC__
#define _PT_(x) x
#else
#define _PT_(x) ()
#endif

int reademall _PT_(( int count ));


/*********************
*  Widget resources  *
*                    *
*********************/

static char *fallbacks[] = {
	"*font:9x15",
	"*reverseVideo:off",
	"*canvas.width:100",
	"*canvas.height:100",
	0
};

/* To get custom resources */

typedef struct {
	Bool help;
	Bool mono;			/* force monochrome display */
	XFontStruct *font;
	Pixel bgc;
	Pixel fgc;
	Pixel hlc;
	Pixel plc[10];
	Bool fill;
	char *label;
	char *file;
	char *title;
	char *conv;
} app_res_t, *app_resp_t;

static app_res_t app_res;

static int minusone = -1;

static XtResource res_list[] = {
	{ "helpme", "XXX", XtRBool, sizeof(Bool),
		XtOffset(app_resp_t, help), XtRString, "off" },
	{ "monochrome", "Monochrome", XtRBool, sizeof(Bool),
		XtOffset(app_resp_t, mono), XtRString, "off" },
	{ "font", "Font", XtRFontStruct, sizeof(XFontStruct*),
		XtOffset(app_resp_t, font), XtRString, XtDefaultFont },
	{ "background", "Background", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, bgc), XtRString, XtDefaultBackground },
	{ "foreground", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, fgc), XtRString, XtDefaultForeground },
	{ "highlight", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, hlc), XtRString, XtDefaultForeground },
	{ "plotColor0", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, plc[0]), XtRInt, &minusone },
	{ "plotColor1", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, plc[1]), XtRInt, &minusone },
	{ "plotColor2", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, plc[2]), XtRInt, &minusone },
	{ "plotColor3", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, plc[3]), XtRInt, &minusone },
	{ "plotColor4", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, plc[4]), XtRInt, &minusone },
	{ "plotColor5", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, plc[5]), XtRInt, &minusone },
	{ "plotColor6", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, plc[6]), XtRInt, &minusone },
	{ "plotColor7", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, plc[7]), XtRInt, &minusone },
	{ "plotColor8", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, plc[8]), XtRInt, &minusone },
	{ "plotColor9", "Foreground", XtRPixel, sizeof(Pixel),
		XtOffset(app_resp_t, plc[9]), XtRInt, &minusone },
	{ "fill", "XXX", XtRBool, sizeof(Bool),
		XtOffset(app_resp_t, fill), XtRString, "off" },
	{ "label", "XXX", XtRString, sizeof(char *),
		XtOffset(app_resp_t, label), XtRString, DEF_LABEL },
	{ "file", "XXX", XtRString, sizeof(char *),
		XtOffset(app_resp_t, file), XtRString, LOGFILE },
	{ "title", "XXX", XtRString, sizeof(char *),
		XtOffset(app_resp_t, title), XtRString, DEF_TITLE },
	{ "conv", "XXX", XtRString, sizeof(char *),
		XtOffset(app_resp_t, conv), XtRString, "n" },
};

static XrmOptionDescRec knownargs[] = {
	{ "-help", ".helpme", XrmoptionNoArg, "on" },
	{ "-mono", ".monochrome", XrmoptionNoArg, "on" },
	{ "+mono", ".monochrome", XrmoptionNoArg, "off" },
	{ "-fn", ".font", XrmoptionSepArg, 0 },
	{ "-hl", ".highlight", XrmoptionSepArg, 0 },
	{ "-plot0", ".plotColor0", XrmoptionSepArg, 0 },
	{ "-plot1", ".plotColor1", XrmoptionSepArg, 0 },
	{ "-plot2", ".plotColor2", XrmoptionSepArg, 0 },
	{ "-plot3", ".plotColor3", XrmoptionSepArg, 0 },
	{ "-plot4", ".plotColor4", XrmoptionSepArg, 0 },
	{ "-plot5", ".plotColor5", XrmoptionSepArg, 0 },
	{ "-plot6", ".plotColor6", XrmoptionSepArg, 0 },
	{ "-plot7", ".plotColor7", XrmoptionSepArg, 0 },
	{ "-plot8", ".plotColor8", XrmoptionSepArg, 0 },
	{ "-plot9", ".plotColor9", XrmoptionSepArg, 0 },
	{ "-fill", ".fill", XrmoptionNoArg, "on" },
	{ "+fill", ".fill", XrmoptionNoArg, "off" },
	{ "-label", ".label", XrmoptionSepArg, 0 },
	{ "-file", ".file", XrmoptionSepArg, 0 },
	{ "-title", ".title", XrmoptionSepArg, 0 },
	{ "-c", ".conv", XrmoptionNoArg, "c" },
	{ "-f", ".conv", XrmoptionNoArg, "f" },
};


/************
*  Globals  *
*           *
************/

static char rcsid[] = "$Id: xtherm.c,v 1.8 1998/09/09 01:25:56 manchek Exp $";
static XtAppContext context = 0;
static Widget topLevel = 0;
static Display *xDisp = 0;
static int xScrn = 0;
static Window xRootW = 0;
static XFontStruct *font_info = 0;
static Widget canvas = 0;
static GC gcs[4];
/*
static XtCallbackRec callback[2] = { { 0, 0 }, { 0, 0 } };
*/
static Arg args[16];
static struct plot *theplot = 0;
static struct dataset **dsets = 0;
static int ndsets = 0;


int
istic(hr, mn)
	int hr;
	int mn;
{
	if (hr >= 9 && hr < 15)
		return mn < 45 ? 1 : 0;
	if (hr < 3 || hr >= 21)
		return mn < 15 ? 1 : 0;
	return mn < 30 ? 1 : 0;
}


/*	draw_graph()
*
*	Draw the whole temperature graph with labels and stuff
*/

int
draw_graph(wgt, plat, nsets, sets)
	Widget wgt;
	struct plot *plat;
	int nsets;
	struct dataset **sets;
{
	Window w = XtWindow(wgt);
	XWindowAttributes wat;
	int wd, ht;					/* of image */
	int oxo, oyo, owd, oht;		/* whole overlay left, top, width, height */
	int gxo, gyo, gwd, ght;		/* graph left, top, width, height */
	int n;
	int i;
	int x, y, y0;
	int vmin, vmax;
	int lhow;
	int ascent, descent;
	int s;
	struct sample sam;
	int afew;
	char buf[200];

/*
	fprintf(stderr, "draw_graph\n");
*/
	XClearWindow(xDisp, w);
	if (!nsets)
		return 0;

	XGetWindowAttributes(xDisp, w, &wat);
	wd = wat.width;
	ht = wat.height;
	oxo = 0;
	oyo = 0;
	owd = wd;
	oht = ht;
	ascent = font_info->ascent;
	descent = font_info->descent;

	lhow = ((plat->label == 'l' || plat->label == 'b') ? 1 : 0)
			+ ((plat->label == 'r' || plat->label == 'b') ? 2 : 0);

	gxo = XTextWidth(font_info, "188", 3);
	switch (lhow) {
	case 1:
		gwd = owd - gxo;
		break;
	case 2:
		gwd = owd - gxo;
		gxo = 0;
		break;
	case 3:
		gwd = owd - 2 * gxo;
		break;
	case 0:
		gwd = owd;
		gxo = 0;
		break;
	}
	gxo += oxo;
	gyo = 3 + ascent / 2;
	if (plat->title && plat->title[0])
		ght = oht - (ascent + descent) / 2 - gyo - (ascent + descent) - 1;
	else
		ght = oht - (ascent + descent) / 2 - gyo - 1;
	gyo += oyo;

	/*
	* get min and max temperatures for all data sets
	*/

	vmin = 200;
	vmax = -200;
	for (s = 0; s < nsets; s++) {
		n = min(gwd, sets[s]->count);
		for (i = 0; i < n; i++) {
			y = sets[s]->samples[sets[s]->count - n + i].val;
			if (y != NOREADING) {
				if (y < vmin)
					vmin = y;
				if (y > vmax)
					vmax = y;
			}
		}
	}
	vmax++;
	if (vmin > vmax)
		vmin = vmax = 0;
	if (vmin < 0)
		vmin = ((vmin - 9) / 10) * 10;
	else
		vmin = (vmin / 10) * 10;
	if (vmax < 0)
		vmax = (vmax / 10) * 10;
	else
		vmax = ((vmax + 9) / 10) * 10;
	if (vmin == vmax)
		vmax += 10;
/*
	fprintf(stderr, "%d to %d\n", vmin, vmax);
*/

	afew = 0;
	do {

	/*
	* draw day numbers and hour scale tic marks
	*/

		n = min(gwd, sets[0]->count);
		for (i = 0; i < n; i++) {
			sam = sets[0]->samples[sets[0]->count - n + i];
			if (sam.mn < 15) {
				if (sam.hr == 0) {
					x = gxo + gwd - n + i;
					XDrawLine(xDisp, w, gcs[AXCOLOR], x, gyo,
							x, gyo + ascent + descent + 1);
				}
				if (sam.hr == 12) {
#ifdef DISPLAY_DAYNAME
					static char *daynames[] = {
						"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat",
					};
					struct tm tm, *ptm;
					time_t ti;
					tm.tm_sec = 0;
					tm.tm_min = sam.mn;
					tm.tm_hour = sam.hr;
					tm.tm_mday = sam.da;
					tm.tm_mon = sam.mo - 1;
					tm.tm_year = (sam.yr < 100 ? sam.yr : sam.yr - 1900);
					tm.tm_wday = 0;
					tm.tm_yday = 0;
					tm.tm_isdst = 0;
					ti = mktime(&tm);
					ptm = localtime(&ti);
					sprintf(buf, "%d/%d %s", ptm->tm_mon + 1, ptm->tm_mday,
						daynames[ptm->tm_wday]);
#else /*DISPLAY_DAYNAME*/
					sprintf(buf, "%d", sam.da);
#endif /*DISPLAY_DAYNAME*/
					x = gxo + gwd - n + i
							- XTextWidth(font_info, buf, strlen(buf)) / 2;
					XDrawString(xDisp, w, gcs[AXCOLOR],
							x, gyo + ascent + 2,
							buf, strlen(buf));
				}
			}
			if (istic(sam.hr, sam.mn)) {
				x = gxo + gwd - n + i;
				XDrawLine(xDisp, w, gcs[AXCOLOR], x, oyo + 1, x, oyo + 3);
			}
		}

	/*
	* draw temperature scale
	*/

		if (lhow)
			for (i = vmin; i <= vmax; i += plat->step) {
				sprintf(buf, "%d", i);
				y = gyo + ght - 1 - ((i - vmin) * (ght - 1)) / (vmax - vmin);
				XDrawLine(xDisp, w, gcs[AXCOLOR], gxo, y, gxo + gwd - 1, y);
				if (lhow & 1) {
					XDrawString(xDisp, w, gcs[AXCOLOR],
							oxo + 2, y + ascent / 2,
							buf, strlen(buf));
				}
				if (lhow & 2) {
					XDrawString(xDisp, w, gcs[AXCOLOR],
							gxo + gwd + 2, y + ascent / 2,
							buf, strlen(buf));
				}
			}

		if (plat->solid && afew)
			break;
		afew++;

	/*
	* draw all temperature plots
	*/

		for (s = 0; s < nsets; s++) {
			n = min(gwd, sets[s]->count);
			y0 = -1;
			for (i = 0; i < n; i++) {
				x = gxo + gwd - n + i;
				sam = sets[s]->samples[sets[s]->count - n + i];
				if (sam.val != NOREADING) {
					y = gyo + ght - 1
							- ((sam.val - vmin) * (ght - 1)) / (vmax - vmin);
					if (plat->solid) {
						XDrawLine(xDisp, w, sets[s]->gc, x, y, x, ht - 1);

					} else {
						if (y0 == -1) {
							XDrawLine(xDisp, w, sets[s]->gc, x, y, x, y);

						} else {
							if (y < y0) {
								XDrawLine(xDisp, w, sets[s]->gc, x, y0, x, y);

							} else {
								XDrawLine(xDisp, w, sets[s]->gc, x - 1, y0, x - 1, y);
								XDrawLine(xDisp, w, sets[s]->gc, x - 1, y, x, y);
							}
						}
					}
					y0 = y;
				} else
					y0 = -1;
			}
		}
	} while (plat->solid);

	if (plat->title && plat->title[0]) {
		XDrawString(xDisp, w, gcs[AXCOLOR],
				oxo + (owd - XTextWidth(font_info, buf, strlen(buf))) / 2,
				oyo + oht - descent - 2,
				plat->title, strlen(plat->title));
	}

	return 0;
}


void
canvas_ev(wgt, cli, xev, rem)
	Widget wgt;
	XtPointer cli;
	XEvent *xev;
	Boolean *rem;
{
/*
	fprintf(stderr, "canvas_ev\n");
*/
	if (xev->type == Expose && ((XExposeEvent*)xev)->count == 0) {
		draw_graph(wgt, theplot, ndsets, dsets);
	}
	if (xev->type == ConfigureNotify) {
		Window w = XtWindow(wgt);
		XWindowAttributes wat;

		XGetWindowAttributes(xDisp, w, &wat);
		if (ndsets > 0 && wat.width > dsets[0]->count)
			reademall((int)wat.width);
		draw_graph(wgt, theplot, ndsets, dsets);
	}
}


void
clock_tic(cli, timer)
	XtPointer cli;
	XtIntervalId *timer;
{
	time_t now;
	struct tm *lt;
	int n;
	Window w = XtWindow((Widget)cli);
	XWindowAttributes wat;

/*
	fprintf(stderr, "clock_tic\n");
*/
	XGetWindowAttributes(xDisp, w, &wat);

	reademall((int)wat.width);
	draw_graph((Widget)cli, theplot, ndsets, dsets);
	now = time(0);
	lt = localtime(&now);
/*
	n = 1000 * (61 - lt->tm_sec);
*/
#ifdef	QUICKTIMERTEST
	n = 5000;
#else
	n = 15 * 60 * 1000;
#endif
	XtAppAddTimeOut(context, n, clock_tic, (XtPointer)canvas);
}


int
main(argc, argv)
	int argc;
	char **argv;
{
	int n;
/*
	Pixmap g50_pm, g75_pm;
*/
	XGCValues xgcv;
/*
	Colormap xCmap;
	XColor colr;
*/
	time_t now;
	struct tm *lt;
	Window w;
	XWindowAttributes wat;
	int mdsets = 0;
	int i;

	n = 0;
	topLevel = XtAppInitialize(&context, "xtherm",
		knownargs, XtNumber(knownargs),
		&argc, argv,
		fallbacks,
		args, n);

	XtGetApplicationResources(topLevel, (caddr_t)&app_res,
		res_list, XtNumber(res_list), 0, 0);

	if (app_res.help || argc < 2) {
/*
		for (i = 1; i < argc; i++)
			if (strcmp("-help", argv[i]))
				fprintf(stderr, "xtherm: unknown option <%s>\n", argv[i]);
*/
		fprintf(stderr,
"usage: xtherm [options] file ...\n\
  option:          resource:       action:\n\
  -/+mono          .monochrome     pretend display is monochrome\n\
  -/+fill          .fill           fill area below graph\n\
  -bg color        .background     set background color\n\
  -fg color        .foreground     set foreground (trace) color\n\
  -hl color        .highlight      set scale color (default fg)\n\
  -plot0..9 color  .plotColor0..9  set individual plot color (default fg)\n\
  -fn font         .font           set font\n\
  -label [lrbn]    .label          label graph on left, right, both, none\n\
  -file name       .file           default log file\n\
  -title string    .title          title label for graph\n\
  -c               .conv c         convert to celsius\n\
  -f               .conv f         convert to fahrenheit\n");
		exit(1);
	}

	xDisp = XtDisplay(topLevel);
	xScrn = DefaultScreen(xDisp);
	xRootW = RootWindow(xDisp, xScrn);
	font_info = app_res.font;

	theplot = TALLOC(1, struct plot, 0);
	theplot->title = app_res.title ? STRALLOC(app_res.title) : 0;
	theplot->label = app_res.label[0];
	theplot->conv = app_res.conv[0];
	theplot->solid = app_res.fill ? 1 : 0;
	theplot->step = 10;

	for (i = 1; i < argc; i++) {
		if (ndsets == mdsets) {
			mdsets = (mdsets * 8) / 5 + 1;
			if (dsets)
				dsets = TREALLOC(dsets, mdsets, struct dataset *);
			else
				dsets = TALLOC(mdsets, struct dataset *, 0);
		}
		dsets[ndsets] = TALLOC(1, struct dataset, 0);
		dsets[ndsets]->count = 0;
		dsets[ndsets]->samples = 0;
		dsets[ndsets]->file = argv[i];
		xgcv.function = GXcopy;
		xgcv.background = app_res.bgc;
		xgcv.foreground = (app_res.plc[ndsets % 10] == minusone)
			? app_res.fgc : app_res.plc[ndsets % 10];
/*
		dsets[ndsets]->px = grcolset ?
				pixs[GRCOLOR] : defpix[ndsets % ndefpix];
*/
		xgcv.font = font_info->fid;
		dsets[ndsets]->gc = XCreateGC(xDisp, xRootW,
				GCBackground|GCForeground|GCFunction|GCFont, &xgcv);

		ndsets++;
	}

	xgcv.function = GXcopy;
	xgcv.background = app_res.bgc;
	xgcv.foreground = app_res.hlc;
	xgcv.font = font_info->fid;
	gcs[AXCOLOR] = XCreateGC(xDisp, xRootW,
		GCBackground|GCForeground|GCFunction|GCFont, &xgcv);

	n = 0;
	canvas = XtCreateManagedWidget("canvas", widgetClass, topLevel, args, n);
	XtRealizeWidget(topLevel);
	XtAddEventHandler(canvas,
		StructureNotifyMask|ExposureMask,
/*
		StructureNotifyMask|ExposureMask|ButtonPressMask|ButtonReleaseMask|KeyPressMask,
*/
		False, canvas_ev, 0);
	now = time(0);
	lt = localtime(&now);
/*
	n = 1000 * (61 - lt->tm_sec);
*/
#ifdef	QUICKTIMERTEST
	n = 5000;
#else
	n = 15 * 60 * 1000;
#endif
	XtAppAddTimeOut(context, n, clock_tic, (XtPointer)canvas);

	w = XtWindow(canvas);
	XGetWindowAttributes(xDisp, w, &wat);
	reademall((int)wat.width);

/*
	fprintf(stderr, "main loop\n");
*/
	XtAppMainLoop(context);
	exit(0);
}


int
reademall(count)
	int count;
{
	int i, j;
	FILE *ff;
	char buf[200];

	if (count < 1)
		count = 1;

	for (i = 0; i < ndsets; i++) {
		if (count > dsets[i]->count) {
			if (dsets[i]->samples)
				MYFREE(dsets[i]->samples);
			dsets[i]->count = count;
			dsets[i]->samples = TALLOC(count, struct sample, 0);
		}

		sprintf(buf, "tail -%d %s", count, dsets[i]->file);
		if ((ff = popen(buf, "r"))) {
/*
			fprintf(stderr, "reademall() %s\n", buf);
*/
			dsets[i]->count = readlogfile(ff, dsets[i]->samples, count);
			if (dsets[i]->count == -1)
				dsets[i]->count = 0;
			fclose(ff);
			if (theplot->conv == 'f')
				for (j = dsets[i]->count; j-- > 0; )
					if (dsets[i]->samples[j].val != NOREADING)
						dsets[i]->samples[j].val =
								(dsets[i]->samples[j].val * 9) / 5 + 32;
			if (theplot->conv == 'c')
				for (j = dsets[i]->count; j-- > 0; )
					if (dsets[i]->samples[j].val != NOREADING)
						dsets[i]->samples[j].val =
								((dsets[i]->samples[j].val - 32) * 5) / 9;
			(void)wait((int *)0);
		} else {
			perror(dsets[i]->file);
		}
	}
	return 0;
}


